import enum
import logging
from dataclasses import dataclass
from typing import cast, Tuple, Union, List, Callable, Dict, TypeVar, Type
from typing_extensions import Literal
from opentrons import types as top_types
from opentrons_shared_data.pipette.types import PipetteChannelType
from opentrons.config import feature_flags

MODULE_LOG = logging.getLogger(__name__)


class MotionChecks(enum.Enum):
    NONE = 0
    LOW = 1
    HIGH = 2
    BOTH = 3


class OT3Mount(enum.Enum):
    LEFT = top_types.Mount.LEFT.value
    RIGHT = top_types.Mount.RIGHT.value
    GRIPPER = enum.auto()

    @classmethod
    def from_mount(
        cls,
        mount: Union[
            top_types.Mount, top_types.MountType, top_types.OT3MountType, "OT3Mount"
        ],
    ) -> "OT3Mount":
        if mount == top_types.Mount.EXTENSION or mount == top_types.MountType.EXTENSION:
            return OT3Mount.GRIPPER
        return cls[mount.name]

    def to_mount(self) -> top_types.Mount:
        if self.value == self.GRIPPER.value:
            return top_types.Mount.EXTENSION
        return top_types.Mount[self.name]

    @classmethod
    def pipette_mounts(cls) -> List["Literal[OT3Mount.LEFT, OT3Mount.RIGHT]"]:
        return [cls.LEFT, cls.RIGHT]


class OT3AxisKind(enum.Enum):
    """An enum of the different kinds of axis we have.

    The machine may have different numbers of specific axes implementing
    each axis kind.
    """

    X = 0
    #: Gantry X axis
    Y = 1
    #: Gantry Y axis
    Z = 2
    #: Z axis (of the left and right)
    P = 3
    #: Plunger axis (of the left and right pipettes)
    Z_G = 4
    #: Gripper Z axis
    Q = 6
    #: High-throughput tip grabbing axis
    OTHER = 6
    #: The internal axes of high throughput pipettes, for instance

    def __str__(self) -> str:
        return self.name

    def is_z_axis(self) -> bool:
        return self in [OT3AxisKind.Z, OT3AxisKind.Z_G]


class Axis(enum.Enum):
    X = 0  # gantry
    Y = 1
    Z_L = 2  # left pipette mount Z
    Z_R = 3  # right pipette mount Z
    Z_G = 4  # gripper mount Z
    P_L = 5  # left pipette plunger
    P_R = 6  # right pipette plunger
    Q = 7  # hi-throughput pipette tiprack grab
    G = 8  # gripper grab

    # OT2 axes' aliases:
    Z = Z_L  # left pipette mount Z
    A = Z_R  # right pipette mount Z
    B = P_L  # left pipette plunger
    C = P_R  # right pipette plunger

    @classmethod
    def by_mount(cls, mount: Union[top_types.Mount, OT3Mount]) -> "Axis":
        bm = {
            top_types.Mount.LEFT: cls.Z_L,
            top_types.Mount.RIGHT: cls.Z_R,
            top_types.Mount.EXTENSION: cls.Z_G,
            OT3Mount.LEFT: cls.Z_L,
            OT3Mount.RIGHT: cls.Z_R,
            OT3Mount.GRIPPER: cls.Z_G,
        }
        return bm[mount]

    @classmethod
    def pipette_axes(cls) -> Tuple["Axis", "Axis"]:
        """The axes which are used for moving plunger motors."""
        return cls.P_L, cls.P_R

    @classmethod
    def mount_axes(cls) -> None:
        """The axes which are used for moving instruments up and down."""
        raise NotImplementedError(
            "`Axis.mount_axes` has been removed. Use `Axis.ot2_mount_axes` or "
            "`Axis.ot3_mount_axes` instead."
        )

    @classmethod
    def ot2_mount_axes(cls) -> Tuple["Axis", "Axis"]:
        """The OT2 axes which are used for moving instruments up and down."""
        # TODO (spp, 2023-07-14): make this a separate function outside of Axis
        # Does this need to be Z_R, Z_L ?
        return cls.Z_L, cls.Z_R

    @classmethod
    def ot3_mount_axes(cls) -> Tuple["Axis", "Axis", "Axis"]:
        """The OT3 axes which are used for moving instruments up and down."""
        # TODO (spp, 2023-07-14): make this a separate function outside of Axis
        return cls.Z_R, cls.Z_L, cls.Z_G

    @classmethod
    def gantry_axes(
        cls,
    ) -> Tuple["Axis", "Axis", "Axis", "Axis", "Axis"]:
        """The axes which are tied to the gantry and require the deck
        calibration transform
        """
        return cls.X, cls.Y, cls.Z_L, cls.Z_R, cls.Z_G

    @classmethod
    def of_main_tool_actuator(cls, mount: Union[top_types.Mount, OT3Mount]) -> "Axis":
        if isinstance(mount, top_types.Mount):
            checked_mount = OT3Mount.from_mount(mount)
        else:
            checked_mount = mount
        pm = {OT3Mount.LEFT: cls.P_L, OT3Mount.RIGHT: cls.P_R, OT3Mount.GRIPPER: cls.G}
        return pm[checked_mount]

    @classmethod
    def to_kind(cls, axis: "Axis") -> OT3AxisKind:
        kind_map: Dict[Axis, OT3AxisKind] = {
            cls.P_L: OT3AxisKind.P,
            cls.P_R: OT3AxisKind.P,
            cls.X: OT3AxisKind.X,
            cls.Y: OT3AxisKind.Y,
            cls.Z_L: OT3AxisKind.Z,
            cls.Z_R: OT3AxisKind.Z,
            cls.Z_G: OT3AxisKind.Z_G,
            cls.Q: OT3AxisKind.Q,
            cls.G: OT3AxisKind.OTHER,
        }
        return kind_map[axis]

    @classmethod
    def of_kind(cls, kind: OT3AxisKind) -> List["Axis"]:
        kind_map: Dict[OT3AxisKind, List[Axis]] = {
            OT3AxisKind.P: [cls.P_R, cls.P_L],
            OT3AxisKind.X: [cls.X],
            OT3AxisKind.Y: [cls.Y],
            OT3AxisKind.Z: [cls.Z_L, cls.Z_R],
            OT3AxisKind.Z_G: [cls.Z_G],
            OT3AxisKind.Q: [cls.Q],
            OT3AxisKind.OTHER: [cls.Q, cls.G],
        }
        return kind_map[kind]

    @classmethod
    def to_ot2_mount(cls, inst: "Axis") -> top_types.Mount:
        # TODO (spp, 2023-07-14): make this a separate function outside of Axis
        return {
            cls.Z: top_types.Mount.LEFT,
            cls.A: top_types.Mount.RIGHT,
            cls.B: top_types.Mount.LEFT,
            cls.C: top_types.Mount.RIGHT,
        }[inst]

    @classmethod
    def to_ot3_mount(cls, inst: "Axis") -> OT3Mount:
        # TODO (spp, 2023-07-14): make this a separate function outside of Axis
        return {
            cls.Z_R: OT3Mount.RIGHT,
            cls.Z_L: OT3Mount.LEFT,
            cls.P_L: OT3Mount.LEFT,
            cls.P_R: OT3Mount.RIGHT,
            cls.Z_G: OT3Mount.GRIPPER,
            cls.G: OT3Mount.GRIPPER,
        }[inst]

    def __str__(self) -> str:
        return self.name

    def of_point(self, point: top_types.Point) -> float:
        if Axis.to_kind(self).is_z_axis():
            return point.z
        elif self == Axis.X:
            return point.x
        elif self == Axis.Y:
            return point.y
        else:
            raise KeyError(self)

    def set_in_point(self, point: top_types.Point, position: float) -> top_types.Point:
        if Axis.to_kind(self).is_z_axis():
            return point._replace(z=position)
        elif self == Axis.X:
            return point._replace(x=position)
        elif self == Axis.Y:
            return point._replace(y=position)
        else:
            raise KeyError(self)

    @classmethod
    def ot2_axes(cls) -> List["Axis"]:
        """Returns only OT2 axes."""
        # TODO (spp, 2023-07-14): make this a separate function outside of Axis
        return [axis for axis in Axis if axis not in [Axis.Z_G, Axis.Q, Axis.G]]

    @classmethod
    def of_plunger(cls, mount: top_types.Mount) -> "Axis":
        """Get plunger axes.

        Same as `of_main_tool_actuator` but for OT2 backwards compatibility.
        """
        return cls.of_main_tool_actuator(mount)

    @classmethod
    def node_axes(cls) -> List["Axis"]:
        """
        Get a list of axes that are backed by flex canbus nodes.
        """
        return [cls.X, cls.Y, cls.Z_L, cls.Z_R, cls.P_L, cls.P_R, cls.Z_G, cls.G]


class SubSystem(enum.Enum):
    """An enumeration of ot3 components.

    This is a complete list of unique firmware nodes in the ot3.
    """

    gantry_x = 0
    gantry_y = 1
    head = 2
    pipette_left = 3
    pipette_right = 4
    gripper = 5
    rear_panel = 6
    motor_controller_board = 7
    hepa_uv = 8

    def __str__(self) -> str:
        return self.name

    @classmethod
    def of_mount(
        cls: "Type[SubSystem]", mount: Union[top_types.Mount, OT3Mount]
    ) -> "Literal[SubSystem.pipette_left, SubSystem.pipette_right, SubSystem.gripper]":
        return cast(
            Literal[SubSystem.pipette_left, SubSystem.pipette_right, SubSystem.gripper],
            {
                top_types.Mount.LEFT: cls.pipette_left,
                top_types.Mount.RIGHT: cls.pipette_right,
                OT3Mount.LEFT: cls.pipette_left,
                OT3Mount.RIGHT: cls.pipette_right,
                OT3Mount.GRIPPER: cls.gripper,
            }[mount],
        )


OT3SubSystem = SubSystem


class PipetteSubType(enum.Enum):
    """Pipette type to map from lower level PipetteType."""

    pipette_single = 1
    pipette_multi = 2
    pipette_96 = 3

    def __str__(self) -> str:
        return self.name

    @classmethod
    def from_channels(cls, channels: PipetteChannelType) -> "PipetteSubType":
        pipette_subtype_lookup = {
            PipetteChannelType.SINGLE_CHANNEL: cls.pipette_single,
            PipetteChannelType.EIGHT_CHANNEL: cls.pipette_multi,
            PipetteChannelType.NINETY_SIX_CHANNEL: cls.pipette_96,
        }
        return pipette_subtype_lookup[channels]


class UpdateState(enum.Enum):
    """Update state to map from lower level FirmwareUpdateStatus"""

    queued = "queued"
    updating = "updating"
    done = "done"

    def __str__(self) -> str:
        return self.value


@dataclass(frozen=True)
class UpdateStatus:
    subsystem: SubSystem
    state: UpdateState
    progress: int


@dataclass
class SubSystemState:
    ok: bool
    current_fw_version: int
    next_fw_version: int
    fw_update_needed: bool
    current_fw_sha: str
    pcba_revision: str
    update_state: Union[UpdateState, None]


BCAxes = Axis  # This doesn't seem to be used. Remove?
AxisMapValue = TypeVar("AxisMapValue")
OT3AxisMap = Dict[Axis, AxisMapValue]


@dataclass
class CurrentConfig:
    hold_current: float
    run_current: float

    def as_tuple(self) -> Tuple[float, float]:
        return self.hold_current, self.run_current


@dataclass(frozen=True)
class MotorStatus:
    motor_ok: bool
    encoder_ok: bool


class DoorState(enum.Enum):
    OPEN = False
    CLOSED = True

    def __str__(self) -> str:
        return self.name.lower()


class EstopState(enum.Enum):
    """Enumerated state machine for the estop status."""

    PHYSICALLY_ENGAGED = enum.auto()
    LOGICALLY_ENGAGED = enum.auto()
    DISENGAGED = enum.auto()
    NOT_PRESENT = enum.auto()


class EstopAttachLocation(enum.Enum):
    """Enumerated estop attach locations."""

    LEFT = enum.auto()
    RIGHT = enum.auto()


class EstopPhysicalStatus(enum.Enum):
    """Possible status of an estop."""

    ENGAGED = enum.auto()
    DISENGAGED = enum.auto()
    NOT_PRESENT = enum.auto()


class HardwareEventType(enum.Enum):
    DOOR_SWITCH_CHANGE = enum.auto()
    ERROR_MESSAGE = enum.auto()
    ESTOP_CHANGE = enum.auto()


@dataclass
class EstopOverallStatus:
    state: EstopState
    left_physical_state: EstopPhysicalStatus
    right_physical_state: EstopPhysicalStatus


@dataclass
class HepaFanState:
    fan_on: bool
    duty_cycle: int


@dataclass
class HepaUVState:
    light_on: bool
    uv_duration_s: int
    remaining_time_s: int


@dataclass(frozen=True)
class DoorStateNotification:
    event: Literal[
        HardwareEventType.DOOR_SWITCH_CHANGE
    ] = HardwareEventType.DOOR_SWITCH_CHANGE
    new_state: DoorState = DoorState.CLOSED


@dataclass(frozen=True)
class EstopStateNotification:
    event: Literal[HardwareEventType.ESTOP_CHANGE] = HardwareEventType.ESTOP_CHANGE
    old_state: EstopState = EstopState.NOT_PRESENT
    new_state: EstopState = EstopState.NOT_PRESENT


@dataclass(frozen=True)
class ErrorMessageNotification:
    message: str
    event: Literal[HardwareEventType.ERROR_MESSAGE] = HardwareEventType.ERROR_MESSAGE


# new event types get new dataclasses
# when we add more event types we add them here
HardwareEvent = Union[
    DoorStateNotification, ErrorMessageNotification, EstopStateNotification
]

HardwareEventHandler = Callable[[HardwareEvent], None]
HardwareEventUnsubscriber = Callable[[], None]


RevisionLiteral = Literal["2.1", "A", "B", "C", "UNKNOWN"]


class BoardRevision(enum.Enum):
    UNKNOWN = enum.auto()
    OG = enum.auto()
    A = enum.auto()
    B = enum.auto()
    C = enum.auto()
    FLEX_B2 = enum.auto()

    @classmethod
    def by_bits(cls, rev_bits: Tuple[bool, bool]) -> "BoardRevision":
        br = {
            (True, True): cls.OG,
            (False, True): cls.A,
            (True, False): cls.B,
            (False, False): cls.C,
        }
        return br[rev_bits]

    def real_name(self) -> Union[RevisionLiteral, Literal["UNKNOWN"]]:
        rn = "2.1" if self.name == "OG" else self.name
        return cast(Union[RevisionLiteral, Literal["UNKNOWN"]], rn)

    def __str__(self) -> str:
        return self.real_name()


class CriticalPoint(enum.Enum):
    """Possibilities for the point to move in a move call.

    The active critical point determines the offsets that are added to the
    gantry position when moving a pipette around.
    """

    MOUNT = enum.auto()
    """
    For legacy reasons, the position of the end of a P300 single. The default
    when no pipette is attached, and used for consistent behavior in certain
    contexts (like change pipette) when a variety of different pipettes might
    be attached.
    """

    NOZZLE = enum.auto()
    """
    The end of the nozzle of a single pipette or the end of the back-most
    nozzle of a multipipette. Only relevant when a pipette is present.
    """

    TIP = enum.auto()
    """
    The end of the tip of a single pipette or the end of the back-most
    tip of a multipipette. Only relevant when a pipette is present and
    a tip with known tip length is attached.
    """

    XY_CENTER = enum.auto()
    """
    Separately from the z component of the critical point, XY_CENTER means
    the critical point under consideration is the XY center of the pipette.
    This changes nothing for single pipettes, but makes multipipettes
    move their centers - so between channels 4 and 5 - to the specified
    point. This is the same as the GRIPPER_JAW_CENTER for grippers.
    """

    INSTRUMENT_XY_CENTER = enum.auto()
    """
    The INSTRUMENT_XY_CENTER means the critical point under consideration is
    the XY center of the entire pipette, regardless of configuration.
    No pipettes, single or multi, will change their instrument center point.
    """

    FRONT_NOZZLE = enum.auto()
    """
    The end of the front-most nozzle of a multipipette with a tip attached.
    Only relevant when a multichannel pipette is present.
    """

    GRIPPER_JAW_CENTER = enum.auto()
    """
    The center of the gripper jaw engagement zone, such that if this critical
    point is moved to the center of a labware the gripper will be ready to
    grip it.
    """

    GRIPPER_FRONT_CALIBRATION_PIN = enum.auto()
    """
    The center of the bottom face of a calibration pin inserted in the gripper's
    front calibration pin slot.
    """

    GRIPPER_REAR_CALIBRATION_PIN = enum.auto()
    """
    The center of the bottom face of a calibration pin inserted in the gripper's
    back calibration pin slot.
    """

    Y_CENTER = enum.auto()
    """
    Y_CENTER means the critical point under consideration is at the same X
    coordinate as the default nozzle point (i.e. TIP | NOZZLE | FRONT_NOZZLE)
    but halfway in between the Y axis bounding box of the pipette - it is the
    XY center of the first column in the pipette. It's really only relevant for
    the 96; it will produce the same position as XY_CENTER on an eight or one
    channel pipette.
    """


class ExecutionState(enum.Enum):
    RUNNING = enum.auto()
    PAUSED = enum.auto()
    CANCELLED = enum.auto()

    def __str__(self) -> str:
        return self.name


class HardwareAction(enum.Enum):
    DROPTIP = enum.auto()
    ASPIRATE = enum.auto()
    DISPENSE = enum.auto()
    BLOWOUT = enum.auto()
    PREPARE_ASPIRATE = enum.auto()
    LIQUID_PROBE = enum.auto()

    def __str__(self) -> str:
        return self.name


class PauseType(enum.Enum):
    PAUSE = 0
    DELAY = 1


class StatusBarState(enum.Enum):
    """Semantic status bar states.

    These mostly correspond to cases listed out in the Flex manual.
    """

    IDLE = enum.auto()
    RUNNING = enum.auto()
    PAUSED = enum.auto()
    HARDWARE_ERROR = enum.auto()
    SOFTWARE_ERROR = enum.auto()
    ERROR_RECOVERY = enum.auto()
    CONFIRMATION = enum.auto()
    RUN_COMPLETED = enum.auto()
    UPDATING = enum.auto()
    ACTIVATION = enum.auto()
    DISCO = enum.auto()
    OFF = enum.auto()

    def transient(self) -> bool:
        return self.value in {
            StatusBarState.CONFIRMATION.value,
            StatusBarState.RUN_COMPLETED.value,
            StatusBarState.ACTIVATION.value,
            StatusBarState.DISCO.value,
        }


@dataclass
class AionotifyEvent:
    flags: enum.EnumMeta
    name: str

    @classmethod
    def build(cls, name: str, flags: List[enum.Enum]) -> "AionotifyEvent":
        # See https://github.com/python/mypy/issues/5317
        # as to why mypy cannot detect that list
        # comprehension or variables cannot be dynamically
        # determined to meet the argument criteria for
        # enums. Hence, the type ignore below.
        flag_list = [f.name for f in flags]
        Flag = enum.Enum("Flag", flag_list)  # type: ignore
        return cls(flags=Flag, name=name)


class GripperJawState(enum.Enum):
    UNHOMED = enum.auto()
    #: the gripper must be homed before it can do anything
    HOMED_READY = enum.auto()
    #: the gripper has been homed and is at its fully-open homed position
    GRIPPING = enum.auto()
    #: the gripper is actively force-control gripping something
    HOLDING = enum.auto()
    #: the gripper is in position-control mode
    STOPPED = enum.auto()
    #: the gripper has been homed before but is stopped now


class InstrumentProbeType(enum.Enum):
    PRIMARY = enum.auto()
    SECONDARY = enum.auto()
    BOTH = enum.auto()


class GripperProbe(enum.Enum):
    FRONT = enum.auto()
    REAR = enum.auto()

    @classmethod
    def to_type(cls, gp: "GripperProbe") -> InstrumentProbeType:
        if gp == cls.FRONT:
            return InstrumentProbeType.SECONDARY
        else:
            return InstrumentProbeType.PRIMARY


class TipStateType(enum.Enum):
    ABSENT = 0
    PRESENT = 1

    def __str__(self) -> str:
        return self.name


@dataclass
class HardwareFeatureFlags:
    """
    Hardware configuration options that can be passed to API instances.
    Some options may not be relevant to every robot.

    These generally map to the feature flag options in the opentrons.config
    module.
    """

    use_old_aspiration_functions: bool = (
        False  # To support pipette backwards compatability
    )
    require_estop: bool = True
    stall_detection_enabled: bool = True
    overpressure_detection_enabled: bool = True

    @classmethod
    def build_from_ff(cls) -> "HardwareFeatureFlags":
        """Build from the feature flags configuration file on disc.

        Note that, if this class is built from the default constructor, the values
        of all of the flags are just the default values instead of the values in the
        feature_flags file or environment variables. Use this constructor to ensure
        the right values are pulled in.
        """
        return HardwareFeatureFlags(
            use_old_aspiration_functions=feature_flags.use_old_aspiration_functions(),
            require_estop=feature_flags.require_estop(),
            stall_detection_enabled=feature_flags.stall_detection_enabled(),
            overpressure_detection_enabled=feature_flags.overpressure_detection_enabled(),
        )


class EarlyLiquidSenseTrigger(RuntimeError):
    """Error raised if sensor threshold reached before minimum probing distance."""

    def __init__(
        self, triggered_at: Dict[Axis, float], min_z_pos: Dict[Axis, float]
    ) -> None:
        """Initialize EarlyLiquidSenseTrigger error."""
        super().__init__(
            f"Liquid threshold triggered early at z={triggered_at}mm, "
            f"minimum z position = {min_z_pos}"
        )


class FailedTipStateCheck(RuntimeError):
    """Error raised if the tip ejector state does not match the expected value."""

    def __init__(
        self, expected_state: TipStateType, actual_state: TipStateType
    ) -> None:
        """Initialize FailedTipStateCheck error."""
        super().__init__(
            f"Expected tip state {expected_state}, but received {actual_state}."
        )



import asyncio
import logging

from opentrons.hardware_control.emulation.module_server import ModuleStatusServer
from opentrons.hardware_control.emulation.proxy import Proxy
from opentrons.hardware_control.emulation.settings import Settings
from opentrons.hardware_control.emulation.types import ModuleType

logger = logging.getLogger(__name__)


class Application:
    """The emulator application."""

    def __init__(self, settings: Settings) -> None:
        """Constructor.

        Args:
            settings: Application settings.
        """
        self._settings = settings
        self._status_server = ModuleStatusServer(settings.module_server)
        self._magdeck = Proxy(
            ModuleType.Magnetic, self._status_server, self._settings.magdeck_proxy
        )
        self._temperature = Proxy(
            ModuleType.Temperature,
            self._status_server,
            self._settings.temperature_proxy,
        )
        self._thermocycler = Proxy(
            ModuleType.Thermocycler,
            self._status_server,
            self._settings.thermocycler_proxy,
        )
        self._heatershaker = Proxy(
            ModuleType.Heatershaker,
            self._status_server,
            self._settings.heatershaker_proxy,
        )

    async def run(self) -> None:
        """Run the application."""
        await asyncio.gather(
            self._status_server.run(),
            self._magdeck.run(),
            self._temperature.run(),
            self._thermocycler.run(),
            self._heatershaker.run(),
        )


if __name__ == "__main__":
    logging.basicConfig(format="%(asctime)s:%(message)s", level=logging.DEBUG)
    s = Settings()
    asyncio.run(Application(settings=s).run())



from typing import Optional, Dict, Any
from opentrons_shared_data.errors.exceptions import (
    MotionPlanningFailureError,
    InvalidInstrumentData,
    RobotInUseError,
)


class OutOfBoundsMove(MotionPlanningFailureError):
    def __init__(self, message: str, detail: Dict[str, Any]):
        super().__init__(message=message, detail=detail)


class InvalidCriticalPoint(MotionPlanningFailureError):
    def __init__(self, cp_name: str, instr: str, message: Optional[str] = None):
        super().__init__(
            message=(message or f"Critical point {cp_name} is invalid for a {instr}."),
            detail={"instrument": instr, "critical point": cp_name},
        )


class InvalidPipetteName(InvalidInstrumentData):
    """Raised for an invalid pipette."""

    def __init__(self, name: int, mount: str) -> None:
        super().__init__(
            message=f"Invalid pipette name key {name} on mount {mount}",
            detail={"mount": mount, "name": str(name)},
        )


class InvalidPipetteModel(InvalidInstrumentData):
    """Raised for a pipette with an unknown model."""

    def __init__(self, name: str, model: str, mount: str) -> None:
        super().__init__(detail={"mount": mount, "name": name, "model": model})


class UpdateOngoingError(RobotInUseError):
    """Error when an update is already happening."""

    def __init__(self, message: str) -> None:
        super().__init__(message=message)



from typing_extensions import Protocol

from .types import ProtocolRobotType


class Identifiable(Protocol[ProtocolRobotType]):
    """Protocol specifying support for hardware identification."""

    def get_robot_type(self) -> ProtocolRobotType:
        """Return the enumerated robot type that this API controls.

        When a caller needs to determine whether an API function is expected
        to be present on a hardware_control instance, it is preferable to check
        with this function rather than check the exact type via `isinstance`.
        """
        ...



from .abstract import AbstractAbsorbanceReaderDriver
from .driver import AbsorbanceReaderDriver
from .simulator import SimulatingDriver
from .hid_protocol import AbsorbanceHidInterface

__all__ = [
    "AbstractAbsorbanceReaderDriver",
    "AbsorbanceReaderDriver",
    "SimulatingDriver",
    "AbsorbanceHidInterface",
]



import json
import logging
import os
from pathlib import Path

from typing import Any, Dict, List, Union, Optional, cast
from typing_extensions import Literal

from . import CONFIG, defaults_ot3, defaults_ot2, gripper_config, feature_flags as ff
from opentrons.hardware_control.types import BoardRevision
from .types import CurrentDict, RobotConfig, AxisDict, OT3Config

log = logging.getLogger(__name__)


def current_for_revision(
    current_dict: CurrentDict, revision: BoardRevision
) -> AxisDict:
    """Pull the appropriate current value for the specified revision."""
    if revision == BoardRevision.UNKNOWN:
        return current_dict.get("2.1", current_dict["default"])
    elif revision.real_name() in current_dict:
        return current_dict[revision.real_name()]  # type: ignore
    else:
        return current_dict["default"]


def build_config(robot_settings: Dict[str, Any]) -> Union[RobotConfig, OT3Config]:
    """Build the appropriate config object for the machine.

    Which config object should be used will be detected from the specified config
    object.
    """
    default_robot_model: Union[Literal["OT-3 Standard"], Literal["OT-2 Standard"]] = (
        "OT-3 Standard" if ff.enable_ot3_hardware_controller() else "OT-2 Standard"
    )
    robot_model = robot_settings.get("model", default_robot_model)
    if robot_model == "OT-3 Standard":
        return build_config_ot3(robot_settings)
    else:
        return build_config_ot2(robot_settings)


def build_config_ot2(robot_settings: Dict[str, Any]) -> RobotConfig:
    """Build an OT2 config object with default values for unspecified elements."""
    return defaults_ot2.build_with_defaults(robot_settings)


def build_config_ot3(robot_settings: Dict[str, Any]) -> OT3Config:
    """Build an OT3 config object with default values for unspecified elements."""
    return defaults_ot3.build_with_defaults(robot_settings)


def config_to_save(config: Union[RobotConfig, OT3Config]) -> Dict[str, Any]:
    """Turn the config into a serializable dictionary."""
    if config.model == "OT-2 Standard":
        return defaults_ot2.serialize(config)
    else:
        return defaults_ot3.serialize(config)


def _load_file() -> Dict[str, Any]:
    settings_file = CONFIG["robot_settings_file"]
    log.debug("Loading robot settings from {}".format(settings_file))
    return _load_json(settings_file) or {}


def load() -> Union[RobotConfig, OT3Config]:
    """Load the appropriate config, if it exists, or build defaults."""
    return build_config(_load_file())


def load_ot2() -> RobotConfig:
    """Load an OT2 config, or build defaults."""
    return build_config_ot2(_load_file())


def load_ot3() -> OT3Config:
    """Load an OT3 config, or build defaults."""
    return build_config_ot3(_load_file())


def save_robot_settings(
    config: Union[RobotConfig, OT3Config],
    rs_filename: Optional[str] = None,
    tag: Optional[str] = None,
) -> Dict[str, Any]:
    """Save the specified config to disk as a JSON file."""
    config_dict = config_to_save(config)
    config_json = json_to_save(config_dict)

    # Save everything else in a different file
    filename = rs_filename or CONFIG["robot_settings_file"]
    if tag:
        root, ext = os.path.splitext(filename)
        filename = "{}-{}{}".format(root, tag, ext)
    _save_config_data(config_json, filename=filename)

    return config_dict


def json_to_save(config: Dict[str, Any]) -> str:
    """Build a JSON object of the config, ready to store on disk."""
    return json.dumps(config, sort_keys=True, indent=4)


def backup_configuration(
    config: Union[RobotConfig, OT3Config], tag: Optional[str] = None
) -> None:
    """Save a backup, tagged by timestamp, of the config."""
    import time

    if not tag:
        tag = str(int(time.time() * 1000))
    save_robot_settings(config, tag=tag)


def get_legacy_gantry_calibration() -> Optional[List[List[float]]]:
    """
    Returns the legacy gantry calibration if exists.

    This should happen only if the new deck calibration file does not exist.
    The legacy calibration should then be migrated to the new format.
    """
    gantry_cal = _load_json(CONFIG["deck_calibration_file"])
    if "gantry_calibration" in gantry_cal:
        return gantry_cal["gantry_calibration"]  # type: ignore[no-any-return]
    else:
        return None


def clear() -> None:
    _clear_file(CONFIG["robot_settings_file"])


def _clear_file(filename: Union[str, Path]) -> None:
    log.debug("Deleting {}".format(filename))
    if os.path.exists(filename):
        os.remove(filename)


# TODO: move to util (write a default load, save JSON function)
def _load_json(filename: Union[str, Path]) -> Dict[str, Any]:
    try:
        with open(filename, "r") as file:
            res = json.load(file)
    except FileNotFoundError:
        log.warning("{0} not found. Loading defaults".format(filename))
        res = {}
    except json.decoder.JSONDecodeError:
        log.warning("{0} is corrupt. Loading defaults".format(filename), exc_info=True)
        res = {}
    return cast(Dict[str, Any], res)


def _save_config_data(data: str, filename: Union[str, Path]) -> None:
    try:
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, "w") as file:
            file.write(data)
            file.flush()
            os.fsync(file.fileno())
    except OSError:
        log.exception("Write failed with exception:")


def default_ot2_deck_calibration() -> List[List[float]]:
    return defaults_ot2.DEFAULT_DECK_CALIBRATION_V2


def default_ot3_deck_calibration() -> List[List[float]]:
    return defaults_ot3.DEFAULT_BELT_ATTITUDE


def default_pipette_offset() -> List[float]:
    if ff.enable_ot3_hardware_controller():
        return defaults_ot3.DEFAULT_PIPETTE_OFFSET
    else:
        return defaults_ot2.DEFAULT_PIPETTE_OFFSET


def default_gripper_calibration_offset() -> List[float]:
    return gripper_config.DEFAULT_GRIPPER_CALIBRATION_OFFSET


def default_module_calibration_offset() -> List[float]:
    return defaults_ot3.DEFAULT_MODULE_OFFSET



import asyncio
import logging
from typing import Dict, Optional, Type

from opentrons.drivers.rpi_drivers.types import USBPort

from ..execution_manager import ExecutionManager

from .types import ModuleDisconnectedCallback, ModuleType, SpeedStatus
from .mod_abc import AbstractModule
from .tempdeck import TempDeck
from .magdeck import MagDeck
from .thermocycler import Thermocycler
from .heater_shaker import HeaterShaker
from .absorbance_reader import AbsorbanceReader


log = logging.getLogger(__name__)


# TODO (lc 05-12-2021) This is pretty gross. We should think
# of a better way to do this.
MODULE_TYPE_BY_NAME = {
    MagDeck.name(): MagDeck.MODULE_TYPE,
    TempDeck.name(): TempDeck.MODULE_TYPE,
    Thermocycler.name(): Thermocycler.MODULE_TYPE,
    HeaterShaker.name(): HeaterShaker.MODULE_TYPE,
    AbsorbanceReader.name(): AbsorbanceReader.MODULE_TYPE,
}

_MODULE_CLS_BY_TYPE: Dict[ModuleType, Type[AbstractModule]] = {
    MagDeck.MODULE_TYPE: MagDeck,
    TempDeck.MODULE_TYPE: TempDeck,
    Thermocycler.MODULE_TYPE: Thermocycler,
    HeaterShaker.MODULE_TYPE: HeaterShaker,
    AbsorbanceReader.MODULE_TYPE: AbsorbanceReader,
}


async def build(
    port: str,
    type: ModuleType,
    simulating: bool,
    usb_port: USBPort,
    hw_control_loop: asyncio.AbstractEventLoop,
    execution_manager: ExecutionManager,
    sim_model: Optional[str] = None,
    sim_serial_number: Optional[str] = None,
    disconnected_callback: ModuleDisconnectedCallback = None,
) -> AbstractModule:
    return await _MODULE_CLS_BY_TYPE[type].build(
        port=port,
        usb_port=usb_port,
        simulating=simulating,
        hw_control_loop=hw_control_loop,
        execution_manager=execution_manager,
        sim_model=sim_model,
        sim_serial_number=sim_serial_number,
        disconnected_callback=disconnected_callback,
    )


async def disable_module(module: AbstractModule) -> None:
    """Async function to deactivate a module immediately, for error recovery."""
    if isinstance(module, HeaterShaker):
        await module.deactivate_heater(must_be_running=False)
        if module.speed_status != SpeedStatus.IDLE:
            await module.deactivate_shaker(must_be_running=False)
    else:
        await module.deactivate(must_be_running=False)



from typing import Sequence, Dict, Optional, Any

from opentrons_shared_data.robot.types import RobotType
from opentrons_shared_data.errors import ErrorCodes
from opentrons_shared_data.errors.exceptions import EnumeratedError

from opentrons.config import feature_flags

from opentrons.protocol_reader.protocol_source import (
    ProtocolConfig,
    PythonProtocolConfig,
    JsonProtocolConfig,
)
from opentrons.protocols.api_support.types import APIVersion


# TODO(mm, 2023-05-10): Deduplicate these constants with
# opentrons.protocol_engine.types.DeckType and consider moving to shared-data.
SHORT_TRASH_DECK = "ot2_short_trash"
STANDARD_OT2_DECK = "ot2_standard"
STANDARD_OT3_DECK = "ot3_standard"


LOAD_FIXED_TRASH_GATE_VERSION_PYTHON = APIVersion(2, 15)
LOAD_FIXED_TRASH_GATE_VERSION_JSON = 7


class NoTrashDefinedError(EnumeratedError):
    """
    Error raised when a protocol attempts to automatically access a trash bin without one being loaded.
    """

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a ProtocolEngineError."""
        super().__init__(
            code=ErrorCodes.GENERAL_ERROR,
            message=message,
            detail=detail,
            wrapping=wrapping,
        )


def should_load_fixed_trash_labware_for_python_protocol(
    api_version: APIVersion,
) -> bool:
    """Whether to automatically load the fixed trash as a labware for a Python protocol at protocol start."""
    return api_version <= LOAD_FIXED_TRASH_GATE_VERSION_PYTHON


def should_load_fixed_trash_area_for_python_protocol(
    api_version: APIVersion, robot_type: RobotType
) -> bool:
    """Whether to automatically load the fixed trash addressable area for OT-2 protocols on 2.16 and above."""
    return (
        api_version > LOAD_FIXED_TRASH_GATE_VERSION_PYTHON
        and robot_type == "OT-2 Standard"
    )


def should_load_fixed_trash(protocol_config: ProtocolConfig) -> bool:
    """Decide whether to automatically load fixed trash labware on the deck based on version."""
    load_fixed_trash = False
    if isinstance(protocol_config, PythonProtocolConfig):
        return should_load_fixed_trash_labware_for_python_protocol(
            protocol_config.api_version
        )
    # TODO(jbl 2023-10-27), when schema v8 is out, use a new deck version field to support fixed trash protocols
    elif isinstance(protocol_config, JsonProtocolConfig):
        load_fixed_trash = (
            protocol_config.schema_version <= LOAD_FIXED_TRASH_GATE_VERSION_JSON
        )

    return load_fixed_trash


def guess_from_global_config() -> str:
    """Return the deck type that the host device physically has.

    This only makes sense when the software is running on a real robot.

    When simulating or analyzing a protocol, especially off-robot, don't use this, because it may
    not match the protocol's declared robot type. Use `for_analysis` instead.
    """
    if feature_flags.enable_ot3_hardware_controller():
        return STANDARD_OT3_DECK
    elif feature_flags.short_fixed_trash():
        return SHORT_TRASH_DECK
    else:
        return STANDARD_OT2_DECK


def for_simulation(robot_type: RobotType) -> str:
    """Return the deck type that should be used for simulating and analyzing a protocol.

    Params:
        robot_type: The robot type that the protocol is meant to run on.
    """
    if robot_type == "OT-2 Standard":
        # OT-2 protocols don't have a way of defining whether they're meant to run on a short-trash
        # or standard deck. So when we're simulating an OT-2 protocol, we need to make an
        # arbitrary choice for which deck type to use.
        return STANDARD_OT2_DECK
    elif robot_type == "OT-3 Standard":
        # OT-3s currently only have a single deck type.
        return STANDARD_OT3_DECK



"""Check a deck layout for conflicts."""
from __future__ import annotations

from dataclasses import dataclass
from typing import List, Mapping, NamedTuple, Optional, Set, Union
from typing_extensions import Final

from opentrons_shared_data.labware.types import LabwareUri
from opentrons_shared_data.robot.types import RobotType
from opentrons.motion_planning.adjacent_slots_getters import (
    get_east_west_slots,
    get_south_slot,
    get_adjacent_slots,
    get_adjacent_staging_slot,
)

from opentrons.types import DeckSlotName, StagingSlotName

_FIXED_TRASH_SLOT: Final[Set[DeckSlotName]] = {
    DeckSlotName.FIXED_TRASH,
    DeckSlotName.SLOT_A3,
}


# The maximum height allowed for items adjacent to a Heater-Shaker in the x-direction.
# This value selected to avoid interference with the Heater-Shaker's labware latch.
# For background, see: https://github.com/Opentrons/opentrons/issues/10316
#
# TODO(mc, 2022-06-16): move this constant to the module definition
HS_MAX_X_ADJACENT_ITEM_HEIGHT = 53.0


# URIs of labware that are allowed to exceed HS_MAX_X_ADJACENT_ITEM_HEIGHT.
# These labware do not take up the full width of the slot
# in the area that would interfere with the labware latch.
# For background, see: https://github.com/Opentrons/opentrons/issues/10316
#
# TODO(mc, 2022-06-16): move this constant to the module definition
HS_ALLOWED_ADJACENT_TALL_LABWARE = [
    LabwareUri("opentrons/opentrons_96_filtertiprack_10ul/1"),
    LabwareUri("opentrons/opentrons_96_filtertiprack_200ul/1"),
    LabwareUri("opentrons/opentrons_96_filtertiprack_20ul/1"),
    LabwareUri("opentrons/opentrons_96_tiprack_10ul/1"),
    LabwareUri("opentrons/opentrons_96_tiprack_20ul/1"),
    LabwareUri("opentrons/opentrons_96_tiprack_300ul/1"),
]


@dataclass
class Labware:
    """A normal labware that directly occupies a slot.

    Do not use this to represent a labware that's loaded atop a module.
    Use one of the module types, instead.
    """

    name_for_errors: str
    highest_z: float
    uri: LabwareUri
    is_fixed_trash: bool


@dataclass
class TrashBin:
    """A non-labware trash bin (loaded via api level 2.16 and above)."""

    name_for_errors: str
    highest_z: float


@dataclass
class _Module:
    name_for_errors: str
    highest_z_including_labware: float


@dataclass
class HeaterShakerModule(_Module):
    """A Heater-Shaker module."""


@dataclass
class MagneticBlockModule(_Module):
    """A Magnetic Block module."""


@dataclass
class ThermocyclerModule(_Module):
    """A Thermocycler module."""

    is_semi_configuration: bool
    """Whether this Thermocycler is loaded in its "semi" configuration.

    In this configuration, it's offset to the left, so it takes up fewer deck slots.
    """


@dataclass
class OtherModule(_Module):
    """A module that's not a Heater-Shaker or Thermocycler."""


DeckItem = Union[
    Labware,
    HeaterShakerModule,
    MagneticBlockModule,
    ThermocyclerModule,
    OtherModule,
    TrashBin,
]


class _NothingAllowed(NamedTuple):
    """Nothing is allowed in this slot."""

    location: Union[DeckSlotName, StagingSlotName]
    source_item: DeckItem
    source_location: Union[DeckSlotName, StagingSlotName]

    def is_allowed(self, item: DeckItem) -> bool:
        return False


class _MaxHeight(NamedTuple):
    """Nothing over a certain height is allowed in this slot."""

    location: DeckSlotName
    source_item: DeckItem
    source_location: DeckSlotName
    max_height: float
    allowed_labware: List[LabwareUri]

    def is_allowed(self, item: DeckItem) -> bool:
        if isinstance(item, Labware):
            if item.uri in self.allowed_labware:
                return True
            else:
                return item.highest_z < self.max_height
        elif isinstance(item, _Module):
            return item.highest_z_including_labware < self.max_height
        elif isinstance(item, TrashBin):
            return item.highest_z < self.max_height


class _NoModule(NamedTuple):
    """No module of any kind is allowed in this slot."""

    location: DeckSlotName
    source_item: DeckItem
    source_location: DeckSlotName

    def is_allowed(self, item: DeckItem) -> bool:
        return not isinstance(item, _Module)


class _NoHeaterShakerModule(NamedTuple):
    """No Heater-Shaker module is allowed in this slot."""

    location: DeckSlotName
    source_item: DeckItem
    source_location: DeckSlotName

    def is_allowed(self, item: DeckItem) -> bool:
        return not isinstance(item, HeaterShakerModule)


_DeckRestriction = Union[
    _NothingAllowed,
    _MaxHeight,
    _NoModule,
    _NoHeaterShakerModule,
]
"""A restriction on what is allowed in a given slot."""


class DeckConflictError(ValueError):
    """Adding an item to the deck would cause a conflict."""


# TODO(mm, 2023-02-16): Taking a single int as the deck location doesn't make sense for
# things that don't fit into a single deck slot, like the Thermocycler.
# Refactor this interface to take a more symbolic location.
def check(
    existing_items: Mapping[Union[DeckSlotName, StagingSlotName], DeckItem],
    new_item: DeckItem,
    new_location: Union[DeckSlotName, StagingSlotName],
    robot_type: RobotType,
) -> None:
    """Check a deck layout for conflicts.

    Args:
        existing_items: Existing items on the deck, assumed to be valid.
        new_item: New item to add to the deck.
        new_location: Location where the new item will be added.
        robot_type: The type of the robot to choose the restriction rules.

    Raises:
        DeckConflictError: Adding this item should not be allowed.
    """
    restrictions: List[_DeckRestriction] = []
    # build restrictions driven by existing items
    for location, item in existing_items.items():
        restrictions += _create_restrictions(
            item=item, location=location, robot_type=robot_type
        )

    # check new item against existing restrictions
    for r in restrictions:
        if r.location == new_location and not r.is_allowed(new_item):
            raise DeckConflictError(
                _create_deck_conflict_error_message(restriction=r, new_item=new_item)
            )

    # check new restrictions required by new item
    # do not interfere with existing items
    new_restrictions = _create_restrictions(
        item=new_item, location=new_location, robot_type=robot_type
    )

    for r in new_restrictions:
        existing_item = existing_items.get(r.location)
        if existing_item is not None and not r.is_allowed(existing_item):
            raise DeckConflictError(
                _create_deck_conflict_error_message(
                    restriction=r,
                    existing_item=existing_item,
                )
            )


def _create_ot2_restrictions(  # noqa: C901
    item: DeckItem, location: Union[DeckSlotName, StagingSlotName]
) -> List[_DeckRestriction]:
    restrictions: List[_DeckRestriction] = []
    if isinstance(location, StagingSlotName):
        raise DeckConflictError(f"OT-2 does not support staging slots ({location.id}).")

    if location not in _FIXED_TRASH_SLOT:
        # Disallow a different item from overlapping this item in this deck slot.
        restrictions.append(
            _NothingAllowed(
                location=location,
                source_item=item,
                source_location=location,
            )
        )

    if _is_ot2_fixed_trash(item):
        # A Heater-Shaker can't safely be placed just south of the fixed trash,
        # because the fixed trash blocks access to the screw that locks the
        # Heater-Shaker onto the deck.
        location_south_of_fixed_trash = get_south_slot(location.as_int())
        if location_south_of_fixed_trash is not None:
            restrictions.append(
                _NoHeaterShakerModule(
                    location=DeckSlotName.from_primitive(location_south_of_fixed_trash),
                    source_item=item,
                    source_location=location,
                )
            )

    if isinstance(item, ThermocyclerModule):
        for covered_location in _ot2_slots_covered_by_thermocycler(item):
            restrictions.append(
                _NothingAllowed(
                    location=covered_location,
                    source_item=item,
                    source_location=location,
                )
            )

    if isinstance(item, HeaterShakerModule):
        for hs_covered_location in get_adjacent_slots(location.as_int()):
            restrictions.append(
                _NoModule(
                    location=DeckSlotName.from_primitive(hs_covered_location),
                    source_item=item,
                    source_location=location,
                )
            )

        for hs_covered_location in get_east_west_slots(location.as_int()):
            restrictions.append(
                _MaxHeight(
                    location=DeckSlotName.from_primitive(hs_covered_location),
                    source_item=item,
                    source_location=location,
                    max_height=HS_MAX_X_ADJACENT_ITEM_HEIGHT,
                    allowed_labware=HS_ALLOWED_ADJACENT_TALL_LABWARE,
                )
            )

    return restrictions


def _create_flex_restrictions(
    item: DeckItem, location: Union[DeckSlotName, StagingSlotName]
) -> List[_DeckRestriction]:
    restrictions: List[_DeckRestriction] = [
        _NothingAllowed(
            location=location,
            source_item=item,
            source_location=location,
        )
    ]

    if isinstance(item, (HeaterShakerModule, OtherModule)):
        if isinstance(location, StagingSlotName):
            raise DeckConflictError(
                "Cannot have a module loaded on a staging area slot."
            )
        adjacent_staging_slot = get_adjacent_staging_slot(location)
        if adjacent_staging_slot is not None:
            # You can't have anything on a staging area slot next to a heater-shaker or
            # temperature module because the module caddy physically blocks you from having
            # that staging area slot installed in the first place.
            restrictions.append(
                _NothingAllowed(
                    location=adjacent_staging_slot,
                    source_item=item,
                    source_location=location,
                )
            )

    elif isinstance(item, ThermocyclerModule):
        for covered_location in _flex_slots_covered_by_thermocycler():
            restrictions.append(
                _NothingAllowed(
                    location=covered_location,
                    source_item=item,
                    source_location=location,
                )
            )

    return restrictions


def _create_restrictions(
    item: DeckItem, location: Union[DeckSlotName, StagingSlotName], robot_type: str
) -> List[_DeckRestriction]:

    if robot_type == "OT-2 Standard":
        return _create_ot2_restrictions(item, location)
    else:
        return _create_flex_restrictions(item, location)


def _create_deck_conflict_error_message(
    restriction: _DeckRestriction,
    new_item: Optional[DeckItem] = None,
    existing_item: Optional[DeckItem] = None,
) -> str:
    assert (
        new_item is not None or existing_item is not None
    ), "Conflict error expects either new_item or existing_item"

    if new_item is not None:
        message = (
            f"{restriction.source_item.name_for_errors}"
            f" in slot {restriction.source_location}"
            f" prevents {new_item.name_for_errors}"
            f" from using slot {restriction.location}."
        )

    elif existing_item is not None:
        message = (
            f"{existing_item.name_for_errors} in slot {restriction.location}"
            f" prevents {restriction.source_item.name_for_errors}"
            f" from using slot {restriction.source_location}."
        )

    return message


def _ot2_slots_covered_by_thermocycler(
    thermocycler: ThermocyclerModule,
) -> Set[DeckSlotName]:
    if thermocycler.is_semi_configuration:
        return {DeckSlotName.SLOT_7, DeckSlotName.SLOT_10}
    else:
        return {
            DeckSlotName.SLOT_7,
            DeckSlotName.SLOT_10,
            DeckSlotName.SLOT_8,
            DeckSlotName.SLOT_11,
        }


def _flex_slots_covered_by_thermocycler() -> Set[DeckSlotName]:
    return {DeckSlotName.SLOT_B1, DeckSlotName.SLOT_A1}


def _is_ot2_fixed_trash(item: DeckItem) -> bool:
    return (isinstance(item, Labware) and item.is_fixed_trash) or isinstance(
        item, TrashBin
    )



from __future__ import annotations
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Union,
    Tuple,
    Mapping,
    NamedTuple,
    TYPE_CHECKING,
)

from typing_extensions import TypeGuard

from opentrons_shared_data.labware.labware_definition import LabwareRole
from opentrons_shared_data.pipette.types import PipetteNameType
from opentrons_shared_data.robot.types import RobotType

from opentrons.protocols.api_support.types import APIVersion, ThermocyclerStep
from opentrons.protocols.api_support.util import APIVersionError
from opentrons.protocols.models import LabwareDefinition
from opentrons.types import Mount, DeckSlotName, StagingSlotName, Location
from opentrons.hardware_control.modules.types import (
    ModuleModel,
    MagneticModuleModel,
    TemperatureModuleModel,
    ThermocyclerModuleModel,
    HeaterShakerModuleModel,
    MagneticBlockModel,
    AbsorbanceReaderModel,
)

from .disposal_locations import TrashBin, WasteChute

if TYPE_CHECKING:
    from .labware import Well


# The first APIVersion where Python protocols can specify deck labels like "D1" instead of "1".
_COORDINATE_DECK_LABEL_VERSION_GATE = APIVersion(2, 15)

# The first APIVersion where Python protocols can specify staging deck slots (e.g. "D4")
_STAGING_DECK_SLOT_VERSION_GATE = APIVersion(2, 16)

# Mapping of public Python Protocol API pipette load names
# to names used by the internal Opentrons system
_PIPETTE_NAMES_MAP = {
    "p10_single": PipetteNameType.P10_SINGLE,
    "p10_multi": PipetteNameType.P10_MULTI,
    "p20_single_gen2": PipetteNameType.P20_SINGLE_GEN2,
    "p20_multi_gen2": PipetteNameType.P20_MULTI_GEN2,
    "p50_single": PipetteNameType.P50_SINGLE,
    "p50_multi": PipetteNameType.P50_MULTI,
    "p300_single": PipetteNameType.P300_SINGLE,
    "p300_multi": PipetteNameType.P300_MULTI,
    "p300_single_gen2": PipetteNameType.P300_SINGLE_GEN2,
    "p300_multi_gen2": PipetteNameType.P300_MULTI_GEN2,
    "p1000_single": PipetteNameType.P1000_SINGLE,
    "p1000_single_gen2": PipetteNameType.P1000_SINGLE_GEN2,
    "flex_1channel_50": PipetteNameType.P50_SINGLE_FLEX,
    "flex_8channel_50": PipetteNameType.P50_MULTI_FLEX,
    "flex_1channel_1000": PipetteNameType.P1000_SINGLE_FLEX,
    "flex_8channel_1000": PipetteNameType.P1000_MULTI_FLEX,
    "flex_96channel_1000": PipetteNameType.P1000_96,
}


class InvalidPipetteMountError(ValueError):
    """An error raised when attempting to load pipettes on an invalid mount."""


class PipetteMountTypeError(TypeError):
    """An error raised when an invalid mount type is used for loading pipettes."""


class LabwareDefinitionIsNotAdapterError(ValueError):
    """An error raised when an adapter is attempted to be loaded as a labware."""


class LabwareDefinitionIsNotLabwareError(ValueError):
    """An error raised when a labware is not loaded using `load_labware`."""


class InvalidTrashBinLocationError(ValueError):
    """An error raised when attempting to load trash bins in invalid slots."""


class InvalidFixtureLocationError(ValueError):
    """An error raised when attempting to load a fixture in an invalid cutout."""


def ensure_mount_for_pipette(
    mount: Union[str, Mount, None], pipette: PipetteNameType
) -> Mount:
    """Ensure that an input value represents a valid mount, and is valid for the given pipette."""
    if pipette == PipetteNameType.P1000_96:
        # Always validate the raw mount input, even if the pipette is a 96-channel and we're not going
        # to use the mount value.
        if mount is not None:
            _ensure_mount(mount)
        # Internal layers treat the 96-channel as being on the left mount.
        return Mount.LEFT
    else:
        if mount is None:
            raise InvalidPipetteMountError(
                f"You must specify a left or right mount to load {pipette.value}."
            )
        else:
            return _ensure_mount(mount)


def _ensure_mount(mount: Union[str, Mount]) -> Mount:
    """Ensure that an input value represents a valid Mount."""
    if mount in [Mount.EXTENSION, "extension"]:
        # This would cause existing protocols that might be iterating over mount types
        # for loading pipettes to raise an error because Mount now includes Extension mount.
        # For example, this would raise error-
        # ```
        #   for i, mount in enumerate(Mount):
        #       pipette[i] = ctx.load_instrument("p300_single", mount)
        # ```
        # But this is a very rare use case and none of the protocols in protocol library
        # or protocols seen/ built by support/ science/ apps engg do this so it might be
        # safe to raise this error now?
        raise InvalidPipetteMountError(
            f"Loading pipettes on {mount} is not allowed."
            f"Use the left or right mounts instead."
        )
    if isinstance(mount, Mount):
        return mount

    if isinstance(mount, str):
        try:
            return Mount[mount.upper()]
        except KeyError as e:
            raise InvalidPipetteMountError(
                "If mount is specified as a string, it must be 'left' or 'right';"
                f" instead, {mount} was given."
            ) from e

    raise PipetteMountTypeError(
        "Instrument mount should be 'left', 'right', or an opentrons.types.Mount;"
        f" instead, {mount} was given."
    )


def ensure_pipette_name(pipette_name: str) -> PipetteNameType:
    """Ensure that an input value represents a valid pipette name."""
    pipette_name = ensure_lowercase_name(pipette_name)

    try:
        return _PIPETTE_NAMES_MAP[pipette_name]
    except KeyError:
        raise ValueError(
            f"Cannot resolve {pipette_name} to pipette, must be given valid pipette name."
        ) from None


# TODO(jbl 11-17-2023) this function's original purpose was ensure a valid deck slot for a given robot type
#   With deck configuration, the shape of this should change to better represent it checking if a deck slot
#   (and maybe any addressable area) being valid for that deck configuration
def ensure_and_convert_deck_slot(
    deck_slot: Union[int, str], api_version: APIVersion, robot_type: RobotType
) -> Union[DeckSlotName, StagingSlotName]:
    """Ensure that a primitive value matches a named deck slot.

    Also, convert the deck slot to match the given `robot_type`.

    Params:
        deck_slot: The primitive value to validate. Valid values are like `5`, `"5"`, or `"C2"`.
        api_version: The Python Protocol API version whose rules to use to validate the value.
            Values like `"C2"` are only supported in newer versions.

    Raises:
        TypeError: If you provide something that's not an `int` or `str`.
        ValueError: If the value does not match a known deck slot.
        APIVersionError: If you provide a value like `"C2"`, but `api_version` is too old.

    Returns:
        A `DeckSlotName` appropriate for the given `robot_type`. For example, given `"5"`,
        this will return `DeckSlotName.SLOT_C2` on a Flex.
    """
    if not isinstance(deck_slot, (int, str)):
        raise TypeError(f"Deck slot must be a string or integer, but got {deck_slot}")

    if str(deck_slot).upper() in {"A4", "B4", "C4", "D4"}:
        if api_version < APIVersion(2, 16):
            raise APIVersionError(
                api_element="Using a staging deck slot",
                until_version=f"{_STAGING_DECK_SLOT_VERSION_GATE}",
                current_version=f"{api_version}",
            )
        # Don't need a try/except since we're already pre-validating this
        parsed_staging_slot = StagingSlotName.from_primitive(str(deck_slot))
        return parsed_staging_slot
    else:
        try:
            parsed_slot = DeckSlotName.from_primitive(deck_slot)
        except ValueError as e:
            raise ValueError(f"'{deck_slot}' is not a valid deck slot") from e
        is_ot2_style = parsed_slot.to_ot2_equivalent() == parsed_slot
        if not is_ot2_style and api_version < _COORDINATE_DECK_LABEL_VERSION_GATE:
            alternative = parsed_slot.to_ot2_equivalent().id
            raise APIVersionError(
                api_element=f"Specifying a deck slot like '{deck_slot}'",
                until_version=f"{_COORDINATE_DECK_LABEL_VERSION_GATE}",
                current_version=f"{api_version}",
                extra_message=f"Increase your protocol's apiLevel, or use slot '{alternative}' instead.",
            )

        return parsed_slot.to_equivalent_for_robot_type(robot_type)


def internal_slot_to_public_string(
    slot_name: Union[DeckSlotName, StagingSlotName], robot_type: RobotType
) -> str:
    """Convert an internal `DeckSlotName` to a user-facing Python Protocol API string.

    This normalizes the string to the robot type's native format, like "5" for OT-2s or "C2" for
    Flexes. This probably won't change anything because the internal `DeckSlotName` should already
    match the robot's native format, but it's nice to have an explicit interface barrier.
    """
    if isinstance(slot_name, DeckSlotName):
        return slot_name.to_equivalent_for_robot_type(robot_type).id
    else:
        # No need to convert staging slot names per robot type, since they only exist on Flex.
        return slot_name.id


def ensure_lowercase_name(name: str) -> str:
    """Ensure that a given name string is all lowercase."""
    if not isinstance(name, str):
        raise TypeError(f"Value must be a string, but got {name}")

    return name.lower()


def ensure_definition_is_adapter(definition: LabwareDefinition) -> None:
    """Ensure that one of the definition's allowed roles is `adapter`."""
    if LabwareRole.adapter not in definition.allowedRoles:
        raise LabwareDefinitionIsNotAdapterError(
            f"Labware {definition.parameters.loadName} is not an adapter."
        )


def ensure_definition_is_labware(definition: LabwareDefinition) -> None:
    """Ensure that one of the definition's allowed roles is `labware` or that that field is empty."""
    if definition.allowedRoles and LabwareRole.labware not in definition.allowedRoles:
        raise LabwareDefinitionIsNotLabwareError(
            f"Labware {definition.parameters.loadName} is not defined as a normal labware."
        )


_MODULE_ALIASES: Dict[str, ModuleModel] = {
    "magdeck": MagneticModuleModel.MAGNETIC_V1,
    "magnetic module": MagneticModuleModel.MAGNETIC_V1,
    "magnetic module gen2": MagneticModuleModel.MAGNETIC_V2,
    "tempdeck": TemperatureModuleModel.TEMPERATURE_V1,
    "temperature module": TemperatureModuleModel.TEMPERATURE_V1,
    "temperature module gen2": TemperatureModuleModel.TEMPERATURE_V2,
    "thermocycler": ThermocyclerModuleModel.THERMOCYCLER_V1,
    "thermocycler module": ThermocyclerModuleModel.THERMOCYCLER_V1,
    "thermocycler module gen2": ThermocyclerModuleModel.THERMOCYCLER_V2,
    # No alias for heater-shaker. Use heater-shaker model name for loading.
}

_MODULE_MODELS: Dict[str, ModuleModel] = {
    "magneticModuleV1": MagneticModuleModel.MAGNETIC_V1,
    "magneticModuleV2": MagneticModuleModel.MAGNETIC_V2,
    "temperatureModuleV1": TemperatureModuleModel.TEMPERATURE_V1,
    "temperatureModuleV2": TemperatureModuleModel.TEMPERATURE_V2,
    "thermocyclerModuleV1": ThermocyclerModuleModel.THERMOCYCLER_V1,
    "thermocyclerModuleV2": ThermocyclerModuleModel.THERMOCYCLER_V2,
    "heaterShakerModuleV1": HeaterShakerModuleModel.HEATER_SHAKER_V1,
    "magneticBlockV1": MagneticBlockModel.MAGNETIC_BLOCK_V1,
    "absorbanceReaderV1": AbsorbanceReaderModel.ABSORBANCE_READER_V1,
}


def ensure_module_model(load_name: str) -> ModuleModel:
    """Ensure that a requested module load name matches a known module model."""
    if not isinstance(load_name, str):
        raise TypeError(f"Module load name must be a string, but got {load_name}")

    model = _MODULE_ALIASES.get(load_name.lower()) or _MODULE_MODELS.get(load_name)

    if model is None:
        valid_names = '", "'.join(_MODULE_ALIASES.keys())
        valid_models = '", "'.join(_MODULE_MODELS.keys())
        raise ValueError(
            f"{load_name} is not a valid module load name.\n"
            f'Valid names (ignoring case): "{valid_names}"\n'
            f'You may also use their exact models: "{valid_models}"'
        )

    return model


def ensure_and_convert_trash_bin_location(
    deck_slot: Union[int, str], api_version: APIVersion, robot_type: RobotType
) -> str:
    """Ensure trash bin load location is valid.

    Also, convert the deck slot to a valid trash bin addressable area.
    """

    if robot_type == "OT-2 Standard":
        raise InvalidTrashBinLocationError("Cannot load trash on OT-2.")

    # map trash bin location to addressable area
    trash_bin_slots = [
        DeckSlotName(slot) for slot in ["A1", "B1", "C1", "D1", "A3", "B3", "C3", "D3"]
    ]
    trash_bin_addressable_areas = [
        "movableTrashA1",
        "movableTrashB1",
        "movableTrashC1",
        "movableTrashD1",
        "movableTrashA3",
        "movableTrashB3",
        "movableTrashC3",
        "movableTrashD3",
    ]
    map_trash_bin_addressable_area = {
        slot: addressable_area
        for slot, addressable_area in zip(trash_bin_slots, trash_bin_addressable_areas)
    }

    slot_name_ot3 = ensure_and_convert_deck_slot(deck_slot, api_version, robot_type)
    if not isinstance(slot_name_ot3, DeckSlotName):
        raise ValueError("Staging areas not permitted for trash bin.")
    if slot_name_ot3 not in trash_bin_slots:
        raise InvalidTrashBinLocationError(
            f"Invalid location for trash bin: {slot_name_ot3}.\n"
            f"Valid slots: Any slot in column 1 or 3."
        )

    return map_trash_bin_addressable_area[slot_name_ot3]


def ensure_hold_time_seconds(
    seconds: Optional[float], minutes: Optional[float]
) -> float:
    """Ensure that hold time is expressed in seconds."""
    if seconds is None:
        seconds = 0
    if minutes is not None:
        seconds += minutes * 60
    return seconds


def ensure_thermocycler_repetition_count(repetitions: int) -> int:
    """Ensure thermocycler repetitions is a positive integer."""
    if repetitions <= 0:
        raise ValueError("repetitions must be a positive integer")
    return repetitions


def ensure_thermocycler_profile_steps(
    steps: List[ThermocyclerStep],
) -> List[ThermocyclerStep]:
    """Ensure thermocycler steps are valid and hold time is expressed in seconds only."""
    validated_steps = []
    for step in steps:
        temperature = step.get("temperature")
        hold_mins = step.get("hold_time_minutes")
        hold_secs = step.get("hold_time_seconds")
        if temperature is None:
            raise ValueError("temperature must be defined for each step in cycle")
        if hold_mins is None and hold_secs is None:
            raise ValueError(
                "either hold_time_minutes or hold_time_seconds must be"
                "defined for each step in cycle"
            )
        validated_seconds = ensure_hold_time_seconds(hold_secs, hold_mins)
        validated_steps.append(
            ThermocyclerStep(
                temperature=temperature, hold_time_seconds=validated_seconds
            )
        )
    return validated_steps


def is_all_integers(items: Sequence[Any]) -> TypeGuard[Sequence[int]]:
    """Check that every item in a list is an integer."""
    return all(isinstance(i, int) for i in items)


def is_all_strings(items: Sequence[Any]) -> TypeGuard[Sequence[str]]:
    """Check that every item in a list is a string."""
    return all(isinstance(i, str) for i in items)


def ensure_valid_labware_offset_vector(
    offset: Mapping[str, float]
) -> Tuple[float, float, float]:
    if not isinstance(offset, dict):
        raise TypeError("Labware offset must be a dictionary.")

    try:
        offsets = (offset["x"], offset["y"], offset["z"])
    except KeyError:
        raise TypeError(
            "Labware offset vector is expected to be a dictionary with"
            " with floating point offset values for all 3 axes."
            " For example: {'x': 1.1, 'y': 2.2, 'z': 3.3}"
        )
    if not all(isinstance(v, (float, int)) for v in offsets):
        raise TypeError("Offset values should be a number (int or float).")
    return offsets


class WellTarget(NamedTuple):
    """A movement target that is a well."""

    well: Well
    location: Optional[Location]
    in_place: bool


class PointTarget(NamedTuple):
    """A movement to coordinates"""

    location: Location
    in_place: bool


class NoLocationError(ValueError):
    """Error representing that no location was supplied."""


class LocationTypeError(TypeError):
    """Error representing that the location supplied is of different expected type."""


ValidTarget = Union[WellTarget, PointTarget, TrashBin, WasteChute]


def validate_location(
    location: Union[Location, Well, TrashBin, WasteChute, None],
    last_location: Optional[Location],
) -> ValidTarget:
    """Validate a given location for a liquid handling command.

    Args:
        location: The input location.
        last_location: The last location accessed by the pipette.

    Returns:
        A `WellTarget` if the input location represents a well.
        A `PointTarget` if the input location is an x, y, z coordinate.

    Raises:
        NoLocationError: The is no input location and no cached loaction.
        LocationTypeError: The location supplied is of unexpected type.
    """
    from .labware import Well

    target_location = location or last_location

    if target_location is None:
        raise NoLocationError()

    if not isinstance(target_location, (Location, Well, TrashBin, WasteChute)):
        raise LocationTypeError(
            f"location should be a Well, Location, TrashBin or WasteChute, but it is {location}"
        )

    if isinstance(target_location, (TrashBin, WasteChute)):
        return target_location

    in_place = target_location == last_location

    if isinstance(target_location, Well):
        return WellTarget(well=target_location, location=None, in_place=in_place)

    _, well = target_location.labware.get_parent_labware_and_well()

    return (
        WellTarget(well=well, location=target_location, in_place=in_place)
        if well is not None
        else PointTarget(location=target_location, in_place=in_place)
    )



"""Input file role analysis."""
from dataclasses import dataclass
from typing import List, Sequence, Union

from .file_identifier import (
    IdentifiedFile,
    IdentifiedJsonMain,
    IdentifiedPythonMain,
    IdentifiedLabwareDefinition,
    IdentifiedData,
)
from .protocol_files_invalid_error import ProtocolFilesInvalidError


@dataclass(frozen=True)
class RoleAnalysis:
    """Role analysis results."""

    main_file: Union[IdentifiedJsonMain, IdentifiedPythonMain]
    labware_files: List[IdentifiedLabwareDefinition]
    data_files: List[IdentifiedData]

    @property
    def all_files(self) -> List[IdentifiedFile]:
        """Return all contained files."""
        return [self.main_file, *self.labware_files, *self.data_files]


class RoleAnalysisError(ProtocolFilesInvalidError):
    """Error raised if the input file list is invalid."""


class RoleAnalyzer:
    """Input file role analysis interface."""

    @staticmethod
    def analyze(files: Sequence[IdentifiedFile]) -> RoleAnalysis:
        """Analyze a set of input files to determine the role that each one fills.

        This validates that there is exactly one main protocol file.
        """
        if len(files) == 0:
            raise RoleAnalysisError(
                message="No files were provided.", detail={"kind": "no-files"}
            )

        main_file_candidates: List[Union[IdentifiedJsonMain, IdentifiedPythonMain]] = []
        labware_files: List[IdentifiedLabwareDefinition] = []
        bundled_data_files: List[IdentifiedData] = []

        for f in files:
            if isinstance(f, (IdentifiedJsonMain, IdentifiedPythonMain)):
                main_file_candidates.append(f)
            elif isinstance(f, IdentifiedLabwareDefinition):
                labware_files.append(f)
            else:
                bundled_data_files.append(f)

        if len(main_file_candidates) == 0:
            if len(files) == 1:
                raise RoleAnalysisError(
                    message=f'"{files[0].original_file.name}" is not a valid protocol file.',
                    detail={"kind": "protocol-does-not-have-main"},
                )
            else:
                file_list = ", ".join(f'"{f.original_file.name}"' for f in files)
                raise RoleAnalysisError(
                    message=f"No valid protocol file found in {file_list}.",
                    detail={"kind": "protocol-does-not-have-main"},
                )
        elif len(main_file_candidates) > 1:
            file_list = ", ".join(
                f'"{f.original_file.name}"' for f in main_file_candidates
            )
            raise RoleAnalysisError(
                message=f"Could not pick single main file from {file_list}.",
                detail={"kind": "multiple-main-candidates"},
            )
        else:
            main_file = main_file_candidates[0]

        return RoleAnalysis(
            main_file=main_file,
            labware_files=labware_files,
            data_files=bundled_data_files,
        )



from typing import List

from typing_extensions import Literal

from pydantic import BaseModel, Field


class ModuleConnection(BaseModel):
    """Model a single module connection."""

    url: str = Field(
        ...,
        description="The url (port) value the module driver should connect to. "
        "For example: socket://host:port",
    )
    module_type: str = Field(
        ..., description="What kind of module this connection emulates."
    )
    identifier: str = Field(..., description="Unique id for this emulator.")


class Message(BaseModel):
    """A message sent to module server clients."""

    status: Literal["connected", "disconnected", "dump"] = Field(
        ...,
        description="`dump` includes a complete list of connected emulators. "
        "`connected` for new connections. `disconnected` for emulators "
        "that have disconnected.  ",
    )
    connections: List[ModuleConnection]



"""Structures to represent changes that commands want to make to engine state."""


import dataclasses
import enum
import typing
from datetime import datetime

from opentrons.hardware_control.nozzle_manager import NozzleMap
from opentrons.protocol_engine.resources import pipette_data_provider
from opentrons.protocol_engine.types import DeckPoint, LabwareLocation, TipGeometry
from opentrons.types import MountType
from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from opentrons_shared_data.pipette.types import PipetteNameType


class _NoChangeEnum(enum.Enum):
    NO_CHANGE = enum.auto()


NO_CHANGE: typing.Final = _NoChangeEnum.NO_CHANGE
"""A sentinel value to indicate that a value shouldn't be changed.

Useful when `None` is semantically unclear or already has some other meaning.
"""


NoChangeType: typing.TypeAlias = typing.Literal[_NoChangeEnum.NO_CHANGE]
"""The type of `NO_CHANGE`, as `NoneType` is to `None`.

Unfortunately, mypy doesn't let us write `Literal[NO_CHANGE]`. Use this instead.
"""


class _ClearEnum(enum.Enum):
    CLEAR = enum.auto()


CLEAR: typing.Final = _ClearEnum.CLEAR
"""A sentinel value to indicate that a value should be cleared.

Useful when `None` is semantically unclear or has some other meaning.
"""


ClearType: typing.TypeAlias = typing.Literal[_ClearEnum.CLEAR]
"""The type of `CLEAR`, as `NoneType` is to `None`.

Unfortunately, mypy doesn't let us write `Literal[CLEAR]`. Use this instead.
"""


@dataclasses.dataclass(frozen=True)
class Well:
    """Designates a well in a labware."""

    labware_id: str
    well_name: str


@dataclasses.dataclass(frozen=True)
class AddressableArea:
    """Designates an addressable area."""

    addressable_area_name: str


@dataclasses.dataclass
class PipetteLocationUpdate:
    """An update to a pipette's location."""

    pipette_id: str
    """The ID of the already-loaded pipette."""

    new_location: Well | AddressableArea | None | NoChangeType
    """The pipette's new logical location.

    Note: `new_location=None` means "change the location to `None` (unknown)",
    not "do not change the location".
    """

    new_deck_point: DeckPoint | NoChangeType


@dataclasses.dataclass
class LabwareLocationUpdate:
    """An update to a labware's location."""

    labware_id: str
    """The ID of the already-loaded labware."""

    new_location: LabwareLocation
    """The labware's new location."""

    offset_id: typing.Optional[str]
    """The ID of the labware's new offset, for its new location."""


@dataclasses.dataclass
class LoadedLabwareUpdate:
    """An update that loads a new labware."""

    labware_id: str
    """The unique ID of the new labware."""

    new_location: LabwareLocation
    """The labware's initial location."""

    offset_id: typing.Optional[str]
    """The ID of the labware's offset."""

    display_name: typing.Optional[str]

    definition: LabwareDefinition


@dataclasses.dataclass
class LoadPipetteUpdate:
    """An update that loads a new pipette.

    NOTE: Currently, if this is provided, a PipetteConfigUpdate must always be
    provided alongside it to fully initialize everything.
    """

    pipette_id: str
    """The unique ID of the new pipette."""

    pipette_name: PipetteNameType
    mount: MountType
    liquid_presence_detection: typing.Optional[bool]


@dataclasses.dataclass
class PipetteConfigUpdate:
    """An update to a pipette's config."""

    pipette_id: str
    """The ID of the already-loaded pipette."""

    # todo(mm, 2024-10-14): Does serial_number belong in LoadPipetteUpdate?
    serial_number: str

    config: pipette_data_provider.LoadedStaticPipetteData


@dataclasses.dataclass
class PipetteNozzleMapUpdate:
    """Update pipette nozzle map."""

    pipette_id: str
    nozzle_map: NozzleMap


@dataclasses.dataclass
class PipetteTipStateUpdate:
    """Update pipette tip state."""

    pipette_id: str
    tip_geometry: typing.Optional[TipGeometry]


@dataclasses.dataclass
class TipsUsedUpdate:
    """Represents an update that marks tips in a tip rack as used."""

    pipette_id: str
    """The pipette that did the tip pickup."""

    labware_id: str

    well_name: str
    """The well that the pipette's primary nozzle targeted.

    Wells in addition to this one will also be marked as used, depending on the
    pipette's nozzle layout.
    """


@dataclasses.dataclass
class LiquidLoadedUpdate:
    """An update from loading a liquid."""

    labware_id: str
    volumes: typing.Dict[str, float]
    last_loaded: datetime


@dataclasses.dataclass
class LiquidProbedUpdate:
    """An update from probing a liquid."""

    labware_id: str
    well_name: str
    last_probed: datetime
    height: float | ClearType
    volume: float | ClearType


@dataclasses.dataclass
class LiquidOperatedUpdate:
    """An update from operating a liquid."""

    labware_id: str
    well_name: str
    volume_added: float | ClearType


@dataclasses.dataclass
class AbsorbanceReaderLidUpdate:
    """An update to an absorbance reader's lid location."""

    module_id: str
    is_lid_on: bool


@dataclasses.dataclass
class StateUpdate:
    """Represents an update to perform on engine state."""

    pipette_location: PipetteLocationUpdate | NoChangeType | ClearType = NO_CHANGE

    loaded_pipette: LoadPipetteUpdate | NoChangeType = NO_CHANGE

    pipette_config: PipetteConfigUpdate | NoChangeType = NO_CHANGE

    pipette_nozzle_map: PipetteNozzleMapUpdate | NoChangeType = NO_CHANGE

    pipette_tip_state: PipetteTipStateUpdate | NoChangeType = NO_CHANGE

    labware_location: LabwareLocationUpdate | NoChangeType = NO_CHANGE

    loaded_labware: LoadedLabwareUpdate | NoChangeType = NO_CHANGE

    tips_used: TipsUsedUpdate | NoChangeType = NO_CHANGE

    liquid_loaded: LiquidLoadedUpdate | NoChangeType = NO_CHANGE

    liquid_probed: LiquidProbedUpdate | NoChangeType = NO_CHANGE

    liquid_operated: LiquidOperatedUpdate | NoChangeType = NO_CHANGE

    absorbance_reader_lid: AbsorbanceReaderLidUpdate | NoChangeType = NO_CHANGE

    # These convenience functions let the caller avoid the boilerplate of constructing a
    # complicated dataclass tree.

    @typing.overload
    def set_pipette_location(
        self,
        *,
        pipette_id: str,
        new_labware_id: str,
        new_well_name: str,
        new_deck_point: DeckPoint,
    ) -> None:
        """Schedule a pipette's location to be set to a well."""

    @typing.overload
    def set_pipette_location(
        self,
        *,
        pipette_id: str,
        new_addressable_area_name: str,
        new_deck_point: DeckPoint,
    ) -> None:
        """Schedule a pipette's location to be set to an addressable area."""
        pass

    def set_pipette_location(  # noqa: D102
        self,
        *,
        pipette_id: str,
        new_labware_id: str | NoChangeType = NO_CHANGE,
        new_well_name: str | NoChangeType = NO_CHANGE,
        new_addressable_area_name: str | NoChangeType = NO_CHANGE,
        new_deck_point: DeckPoint,
    ) -> None:
        if new_addressable_area_name != NO_CHANGE:
            self.pipette_location = PipetteLocationUpdate(
                pipette_id=pipette_id,
                new_location=AddressableArea(
                    addressable_area_name=new_addressable_area_name
                ),
                new_deck_point=new_deck_point,
            )
        else:
            # These asserts should always pass because of the overloads.
            assert new_labware_id != NO_CHANGE
            assert new_well_name != NO_CHANGE

            self.pipette_location = PipetteLocationUpdate(
                pipette_id=pipette_id,
                new_location=Well(labware_id=new_labware_id, well_name=new_well_name),
                new_deck_point=new_deck_point,
            )

    def clear_all_pipette_locations(self) -> None:
        """Mark all pipettes as having an unknown location."""
        self.pipette_location = CLEAR

    def set_labware_location(
        self,
        *,
        labware_id: str,
        new_location: LabwareLocation,
        new_offset_id: str | None,
    ) -> None:
        """Set a labware's location. See `LabwareLocationUpdate`."""
        self.labware_location = LabwareLocationUpdate(
            labware_id=labware_id,
            new_location=new_location,
            offset_id=new_offset_id,
        )

    def set_loaded_labware(
        self,
        definition: LabwareDefinition,
        labware_id: str,
        offset_id: typing.Optional[str],
        display_name: typing.Optional[str],
        location: LabwareLocation,
    ) -> None:
        """Add a new labware to state. See `LoadedLabwareUpdate`."""
        self.loaded_labware = LoadedLabwareUpdate(
            definition=definition,
            labware_id=labware_id,
            offset_id=offset_id,
            new_location=location,
            display_name=display_name,
        )

    def set_load_pipette(
        self,
        pipette_id: str,
        pipette_name: PipetteNameType,
        mount: MountType,
        liquid_presence_detection: typing.Optional[bool],
    ) -> None:
        """Add a new pipette to state. See `LoadPipetteUpdate`."""
        self.loaded_pipette = LoadPipetteUpdate(
            pipette_id=pipette_id,
            pipette_name=pipette_name,
            mount=mount,
            liquid_presence_detection=liquid_presence_detection,
        )

    def update_pipette_config(
        self,
        pipette_id: str,
        config: pipette_data_provider.LoadedStaticPipetteData,
        serial_number: str,
    ) -> None:
        """Update a pipette's config. See `PipetteConfigUpdate`."""
        self.pipette_config = PipetteConfigUpdate(
            pipette_id=pipette_id, config=config, serial_number=serial_number
        )

    def update_pipette_nozzle(self, pipette_id: str, nozzle_map: NozzleMap) -> None:
        """Update a pipette's nozzle map. See `PipetteNozzleMapUpdate`."""
        self.pipette_nozzle_map = PipetteNozzleMapUpdate(
            pipette_id=pipette_id, nozzle_map=nozzle_map
        )

    def update_pipette_tip_state(
        self, pipette_id: str, tip_geometry: typing.Optional[TipGeometry]
    ) -> None:
        """Update a pipette's tip state. See `PipetteTipStateUpdate`."""
        self.pipette_tip_state = PipetteTipStateUpdate(
            pipette_id=pipette_id, tip_geometry=tip_geometry
        )

    def mark_tips_as_used(
        self, pipette_id: str, labware_id: str, well_name: str
    ) -> None:
        """Mark tips in a tip rack as used. See `TipsUsedUpdate`."""
        self.tips_used = TipsUsedUpdate(
            pipette_id=pipette_id, labware_id=labware_id, well_name=well_name
        )

    def set_liquid_loaded(
        self,
        labware_id: str,
        volumes: typing.Dict[str, float],
        last_loaded: datetime,
    ) -> None:
        """Add liquid volumes to well state. See `LoadLiquidUpdate`."""
        self.liquid_loaded = LiquidLoadedUpdate(
            labware_id=labware_id,
            volumes=volumes,
            last_loaded=last_loaded,
        )

    def set_liquid_probed(
        self,
        labware_id: str,
        well_name: str,
        last_probed: datetime,
        height: float | ClearType,
        volume: float | ClearType,
    ) -> None:
        """Add a liquid height and volume to well state. See `ProbeLiquidUpdate`."""
        self.liquid_probed = LiquidProbedUpdate(
            labware_id=labware_id,
            well_name=well_name,
            height=height,
            volume=volume,
            last_probed=last_probed,
        )

    def set_liquid_operated(
        self, labware_id: str, well_name: str, volume_added: float | ClearType
    ) -> None:
        """Update liquid volumes in well state. See `OperateLiquidUpdate`."""
        self.liquid_operated = LiquidOperatedUpdate(
            labware_id=labware_id,
            well_name=well_name,
            volume_added=volume_added,
        )

    def set_absorbance_reader_lid(self, module_id: str, is_lid_on: bool) -> None:
        """Update an absorbance reader's lid location. See `AbsorbanceReaderLidUpdate`."""
        self.absorbance_reader_lid = AbsorbanceReaderLidUpdate(
            module_id=module_id, is_lid_on=is_lid_on
        )



from typing import List, Optional

from opentrons.calibration_storage import helpers
from opentrons.protocols.geometry.labware_geometry import LabwareGeometry
from opentrons.protocols.api_support.tip_tracker import TipTracker

from opentrons.types import DeckSlotName, Location, Point
from opentrons.hardware_control.nozzle_manager import NozzleMap
from opentrons_shared_data.labware.types import LabwareParameters, LabwareDefinition

from ..labware import AbstractLabware, LabwareLoadParams
from .legacy_well_core import LegacyWellCore
from .well_geometry import WellGeometry


# URIs of labware whose definitions accidentally specify an engage height
# in units of half-millimeters instead of millimeters.
_MAGDECK_HALF_MM_LABWARE = {
    "opentrons/biorad_96_wellplate_200ul_pcr/1",
    "opentrons/nest_96_wellplate_100ul_pcr_full_skirt/1",
    "opentrons/usascientific_96_wellplate_2.4ml_deep/1",
}


class LegacyLabwareCore(AbstractLabware[LegacyWellCore]):
    """Labware implementation core based on legacy PAPIv2 behavior.

    Args:
        definition: The labware definition, as a dict.
        parent: The location of the labware's origin point.
            A labware's origin is its front and left most corner.
            Usually a slot or a module location.
        label: Optional display label.
    """

    def __init__(
        self,
        definition: LabwareDefinition,
        parent: Location,
        label: Optional[str] = None,
    ) -> None:
        self._label: Optional[str] = None
        if label:
            dn = self._label = label
            self._name = dn
        else:
            dn = definition["metadata"]["displayName"]
            self._name = definition["parameters"]["loadName"]

        self._display_name = f"{dn} on {str(parent.labware)}"
        # Directly from definition
        self._well_definition = definition["wells"]
        self._parameters = definition["parameters"]
        self._definition = definition

        self._geometry = LabwareGeometry(definition, parent)
        self._calibrated_offset = Point(
            x=self._geometry.offset.x,
            y=self._geometry.offset.y,
            z=self._geometry.offset.z,
        )

        # flatten list of list of well names.
        self._wells_by_name = {
            well_name: LegacyWellCore(
                well_geometry=WellGeometry(
                    well_props=self._well_definition[well_name],
                    parent_point=self._calibrated_offset,
                    parent_object=self,
                ),
                display_name=f"{well_name} of {self._display_name}",
                has_tip=self.is_tip_rack(),
                name=well_name,
            )
            for column in self.get_well_columns()
            for well_name in column
        }

        self._tip_tracker = TipTracker(
            columns=[
                [self._wells_by_name[well_name] for well_name in column]
                for column in self.get_well_columns()
            ]
        )

    def get_uri(self) -> str:
        return helpers.uri_from_definition(self._definition)

    def get_load_params(self) -> LabwareLoadParams:
        return LabwareLoadParams(
            namespace=self._definition["namespace"],
            load_name=self._definition["parameters"]["loadName"],
            version=self._definition["version"],
        )

    def get_display_name(self) -> str:
        return self._display_name

    def get_user_display_name(self) -> Optional[str]:
        return self._label

    def get_name(self) -> str:
        return self._name

    def set_name(self, new_name: str) -> None:
        self._name = new_name

    def get_definition(self) -> LabwareDefinition:
        return self._definition

    def get_parameters(self) -> LabwareParameters:
        return self._parameters

    def get_quirks(self) -> List[str]:
        return self._parameters.get("quirks", [])

    def set_calibration(self, delta: Point) -> None:
        self._calibrated_offset = Point(
            x=self._geometry.offset.x + delta.x,
            y=self._geometry.offset.y + delta.y,
            z=self._geometry.offset.z + delta.z,
        )

        # Rebuild well geometry with new offset
        for well_name, well_core in self._wells_by_name.items():
            well_core.geometry = WellGeometry(
                well_props=self._well_definition[well_name],
                parent_point=self._calibrated_offset,
                parent_object=self,
            )

    def get_calibrated_offset(self) -> Point:
        return self._calibrated_offset

    def is_tip_rack(self) -> bool:
        return self._parameters["isTiprack"]

    def is_adapter(self) -> bool:
        return False  # Adapters were introduced in v2.15 and not supported in legacy protocols

    def is_lid(self) -> bool:
        return (
            False  # Lids were introduced in v2.21 and not supported in legacy protocols
        )

    def is_fixed_trash(self) -> bool:
        """Whether the labware is fixed trash."""
        return "fixedTrash" in self.get_quirks()

    def get_tip_length(self) -> float:
        return self._parameters["tipLength"]

    def set_tip_length(self, length: float) -> None:
        self._parameters["tipLength"] = length

    def reset_tips(self) -> None:
        if self.is_tip_rack():
            for well in self._wells_by_name.values():
                well.set_has_tip(True)

    def get_next_tip(
        self,
        num_tips: int,
        starting_tip: Optional[LegacyWellCore],
        nozzle_map: Optional[NozzleMap],
    ) -> Optional[str]:
        if nozzle_map is not None:
            raise ValueError(
                "Nozzle Map cannot be provided to calls for next tip in legacy protocols."
            )
        next_well = self._tip_tracker.next_tip(num_tips, starting_tip)
        return next_well.get_name() if next_well else None

    def get_tip_tracker(self) -> TipTracker:
        return self._tip_tracker

    def get_well_columns(self) -> List[List[str]]:
        """Get the all well names, organized by column, from the labware's definition."""
        return self._definition["ordering"]

    def get_geometry(self) -> LabwareGeometry:
        return self._geometry

    @property
    def highest_z(self) -> float:
        return self._geometry.z_dimension + self._calibrated_offset.z

    @property
    def load_name(self) -> str:
        return self._parameters["loadName"]

    # TODO(mc, 2022-09-26): codify "from labware's base" in definition schema
    # https://opentrons.atlassian.net/browse/RSS-110
    def get_default_magnet_engage_height(
        self, preserve_half_mm: bool = False
    ) -> Optional[float]:
        """Get the labware's default magnet engage height, if defined.

        Value returned is in real millimeters from the labware's base,
        unless `preserve_half_mm` is used, in which case
        some definitions will return half-millimeters.
        """
        is_compatible = self._parameters.get("isMagneticModuleCompatible", False)
        default_engage_height = self._parameters.get("magneticModuleEngageHeight")

        if not is_compatible or default_engage_height is None:
            return None

        if self.get_uri() in _MAGDECK_HALF_MM_LABWARE and not preserve_half_mm:
            # TODO(mc, 2022-09-26): this value likely _also_ needs a few mm subtracted
            # https://opentrons.atlassian.net/browse/RSS-111
            return default_engage_height / 2.0

        return default_engage_height

    def get_well_core(self, well_name: str) -> LegacyWellCore:
        return self._wells_by_name[well_name]

    def get_deck_slot(self) -> Optional[DeckSlotName]:
        """Get the deck slot the labware is in, if in a deck slot."""
        slot = self._geometry.parent.labware.first_parent()
        return DeckSlotName.from_primitive(slot) if slot is not None else None



"""Verify tip presence command request, result and implementation models."""
from __future__ import annotations

from pydantic import Field, BaseModel
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from .pipetting_common import PipetteIdMixin
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence

from ..types import TipPresenceStatus, InstrumentSensorId

if TYPE_CHECKING:
    from ..execution import TipHandler


VerifyTipPresenceCommandType = Literal["verifyTipPresence"]


class VerifyTipPresenceParams(PipetteIdMixin):
    """Payload required for a VerifyTipPresence command."""

    expectedState: TipPresenceStatus = Field(
        ..., description="The expected tip presence status on the pipette."
    )
    followSingularSensor: Optional[InstrumentSensorId] = Field(
        default=None, description="The sensor id to follow if the other can be ignored."
    )


class VerifyTipPresenceResult(BaseModel):
    """Result data from the execution of a VerifyTipPresence command."""

    pass


class VerifyTipPresenceImplementation(
    AbstractCommandImpl[VerifyTipPresenceParams, SuccessData[VerifyTipPresenceResult]]
):
    """VerifyTipPresence command implementation."""

    def __init__(
        self,
        tip_handler: TipHandler,
        **kwargs: object,
    ) -> None:
        self._tip_handler = tip_handler

    async def execute(
        self, params: VerifyTipPresenceParams
    ) -> SuccessData[VerifyTipPresenceResult]:
        """Verify if tip presence is as expected for the requested pipette."""
        pipette_id = params.pipetteId
        expected_state = params.expectedState
        follow_singular_sensor = (
            InstrumentSensorId.to_instrument_probe_type(params.followSingularSensor)
            if params.followSingularSensor
            else None
        )

        await self._tip_handler.verify_tip_presence(
            pipette_id=pipette_id,
            expected=expected_state,
            follow_singular_sensor=follow_singular_sensor,
        )

        return SuccessData(
            public=VerifyTipPresenceResult(),
        )


class VerifyTipPresence(
    BaseCommand[VerifyTipPresenceParams, VerifyTipPresenceResult, ErrorOccurrence]
):
    """VerifyTipPresence command model."""

    commandType: VerifyTipPresenceCommandType = "verifyTipPresence"
    params: VerifyTipPresenceParams
    result: Optional[VerifyTipPresenceResult]

    _ImplementationCls: Type[
        VerifyTipPresenceImplementation
    ] = VerifyTipPresenceImplementation


class VerifyTipPresenceCreate(BaseCommandCreate[VerifyTipPresenceParams]):
    """VerifyTipPresence command creation request model."""

    commandType: VerifyTipPresenceCommandType = "verifyTipPresence"
    params: VerifyTipPresenceParams

    _CommandCls: Type[VerifyTipPresence] = VerifyTipPresence



"""The liquidProbe and tryLiquidProbe commands."""

from __future__ import annotations
from typing import TYPE_CHECKING, NamedTuple, Optional, Type, Union
from typing_extensions import Literal

from pydantic import Field

from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.errors.exceptions import (
    MustHomeError,
    PipetteNotReadyToAspirateError,
    TipNotEmptyError,
    IncompleteLabwareDefinitionError,
)
from opentrons.types import MountType
from opentrons_shared_data.errors.exceptions import (
    PipetteLiquidNotFoundError,
)

from ..types import DeckPoint
from .pipetting_common import (
    LiquidNotFoundError,
    PipetteIdMixin,
    WellLocationMixin,
    DestinationPositionResult,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)

from ..errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from ..execution import MovementHandler, PipettingHandler
    from ..resources import ModelUtils
    from ..state.state import StateView


LiquidProbeCommandType = Literal["liquidProbe"]
TryLiquidProbeCommandType = Literal["tryLiquidProbe"]


# Both command variants should have identical parameters.
# But we need two separate parameter model classes because
# `command_unions.CREATE_TYPES_BY_PARAMS_TYPE` needs to be a 1:1 mapping.
class _CommonParams(PipetteIdMixin, WellLocationMixin):
    pass


class LiquidProbeParams(_CommonParams):
    """Parameters required for a `liquidProbe` command."""

    pass


class TryLiquidProbeParams(_CommonParams):
    """Parameters required for a `tryLiquidProbe` command."""

    pass


class LiquidProbeResult(DestinationPositionResult):
    """Result data from the execution of a `liquidProbe` command."""

    z_position: float = Field(
        ..., description="The Z coordinate, in mm, of the found liquid in deck space."
    )
    # New fields should use camelCase. z_position is snake_case for historical reasons.


class TryLiquidProbeResult(DestinationPositionResult):
    """Result data from the execution of a `tryLiquidProbe` command."""

    z_position: Optional[float] = Field(
        ...,
        description=(
            "The Z coordinate, in mm, of the found liquid in deck space."
            " If no liquid was found, `null` or omitted."
        ),
    )


_LiquidProbeExecuteReturn = Union[
    SuccessData[LiquidProbeResult],
    DefinedErrorData[LiquidNotFoundError],
]
_TryLiquidProbeExecuteReturn = SuccessData[TryLiquidProbeResult]


class _ExecuteCommonResult(NamedTuple):
    # If the probe succeeded, the z_pos that it returned.
    # Or, if the probe found no liquid, the error representing that,
    # so calling code can propagate those details up.
    z_pos_or_error: float | PipetteLiquidNotFoundError

    state_update: update_types.StateUpdate
    deck_point: DeckPoint


async def _execute_common(
    state_view: StateView,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    params: _CommonParams,
) -> _ExecuteCommonResult:
    pipette_id = params.pipetteId
    labware_id = params.labwareId
    well_name = params.wellName

    state_update = update_types.StateUpdate()

    # May raise TipNotAttachedError.
    aspirated_volume = state_view.pipettes.get_aspirated_volume(pipette_id)

    if aspirated_volume is None:
        # Theoretically, we could avoid raising an error by automatically preparing
        # to aspirate above the well like AspirateImplementation does. However, the
        # only way for this to happen is if someone tries to do a liquid probe with
        # a tip that's previously held liquid, which they should avoid anyway.
        raise PipetteNotReadyToAspirateError(
            "The pipette cannot probe liquid because of a previous blow out."
            " The plunger must be reset while the tip is somewhere away from liquid."
        )
    elif aspirated_volume != 0:
        raise TipNotEmptyError(
            message="The pipette cannot probe for liquid when the tip has liquid in it."
        )

    if await movement.check_for_valid_position(mount=MountType.LEFT) is False:
        raise MustHomeError(
            message="Current position of pipette is invalid. Please home."
        )

    # liquid_probe process start position
    position = await movement.move_to_well(
        pipette_id=pipette_id,
        labware_id=labware_id,
        well_name=well_name,
        well_location=params.wellLocation,
    )
    deck_point = DeckPoint.construct(x=position.x, y=position.y, z=position.z)
    state_update.set_pipette_location(
        pipette_id=pipette_id,
        new_labware_id=labware_id,
        new_well_name=well_name,
        new_deck_point=deck_point,
    )

    try:
        z_pos = await pipetting.liquid_probe_in_place(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=params.wellLocation,
        )
    except PipetteLiquidNotFoundError as exception:
        return _ExecuteCommonResult(
            z_pos_or_error=exception, state_update=state_update, deck_point=deck_point
        )
    else:
        return _ExecuteCommonResult(
            z_pos_or_error=z_pos, state_update=state_update, deck_point=deck_point
        )


class LiquidProbeImplementation(
    AbstractCommandImpl[LiquidProbeParams, _LiquidProbeExecuteReturn]
):
    """The implementation of a `liquidProbe` command."""

    def __init__(
        self,
        state_view: StateView,
        movement: MovementHandler,
        pipetting: PipettingHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._movement = movement
        self._pipetting = pipetting
        self._model_utils = model_utils

    async def execute(self, params: _CommonParams) -> _LiquidProbeExecuteReturn:
        """Move to and liquid probe the requested well.

        Return the z-position of the found liquid.
        If no liquid is found, return a LiquidNotFoundError as a defined error.

        Raises:
            TipNotAttachedError: as an undefined error, if there is not tip attached to
                the pipette.
            TipNotEmptyError: as an undefined error, if the tip starts with liquid
                in it.
            PipetteNotReadyToAspirateError: as an undefined error, if the plunger is not
                in a safe position to do the liquid probe.
            MustHomeError: as an undefined error, if the plunger is not in a valid
                position.
        """
        z_pos_or_error, state_update, deck_point = await _execute_common(
            self._state_view, self._movement, self._pipetting, params
        )
        if isinstance(z_pos_or_error, PipetteLiquidNotFoundError):
            state_update.set_liquid_probed(
                labware_id=params.labwareId,
                well_name=params.wellName,
                height=update_types.CLEAR,
                volume=update_types.CLEAR,
                last_probed=self._model_utils.get_timestamp(),
            )
            return DefinedErrorData(
                public=LiquidNotFoundError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=z_pos_or_error,
                        )
                    ],
                ),
                state_update=state_update,
            )
        else:
            try:
                well_volume: float | update_types.ClearType = (
                    self._state_view.geometry.get_well_volume_at_height(
                        labware_id=params.labwareId,
                        well_name=params.wellName,
                        height=z_pos_or_error,
                    )
                )
            except IncompleteLabwareDefinitionError:
                well_volume = update_types.CLEAR
            state_update.set_liquid_probed(
                labware_id=params.labwareId,
                well_name=params.wellName,
                height=z_pos_or_error,
                volume=well_volume,
                last_probed=self._model_utils.get_timestamp(),
            )
            return SuccessData(
                public=LiquidProbeResult(
                    z_position=z_pos_or_error, position=deck_point
                ),
                state_update=state_update,
            )


class TryLiquidProbeImplementation(
    AbstractCommandImpl[TryLiquidProbeParams, _TryLiquidProbeExecuteReturn]
):
    """The implementation of a `tryLiquidProbe` command."""

    def __init__(
        self,
        state_view: StateView,
        movement: MovementHandler,
        pipetting: PipettingHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._movement = movement
        self._pipetting = pipetting
        self._model_utils = model_utils

    async def execute(self, params: _CommonParams) -> _TryLiquidProbeExecuteReturn:
        """Execute a `tryLiquidProbe` command.

        `tryLiquidProbe` is identical to `liquidProbe`, except that if no liquid is
        found, `tryLiquidProbe` returns a success result with `z_position=null` instead
        of a defined error.
        """
        z_pos_or_error, state_update, deck_point = await _execute_common(
            self._state_view, self._movement, self._pipetting, params
        )

        if isinstance(z_pos_or_error, PipetteLiquidNotFoundError):
            z_pos = None
            well_volume: float | update_types.ClearType = update_types.CLEAR
        else:
            z_pos = z_pos_or_error
            try:
                well_volume = self._state_view.geometry.get_well_volume_at_height(
                    labware_id=params.labwareId, well_name=params.wellName, height=z_pos
                )
            except IncompleteLabwareDefinitionError:
                well_volume = update_types.CLEAR

        state_update.set_liquid_probed(
            labware_id=params.labwareId,
            well_name=params.wellName,
            height=z_pos if z_pos is not None else update_types.CLEAR,
            volume=well_volume,
            last_probed=self._model_utils.get_timestamp(),
        )

        return SuccessData(
            public=TryLiquidProbeResult(
                z_position=z_pos,
                position=deck_point,
            ),
            state_update=state_update,
        )


class LiquidProbe(
    BaseCommand[LiquidProbeParams, LiquidProbeResult, LiquidNotFoundError]
):
    """The model for a full `liquidProbe` command."""

    commandType: LiquidProbeCommandType = "liquidProbe"
    params: LiquidProbeParams
    result: Optional[LiquidProbeResult]

    _ImplementationCls: Type[LiquidProbeImplementation] = LiquidProbeImplementation


class TryLiquidProbe(
    BaseCommand[TryLiquidProbeParams, TryLiquidProbeResult, ErrorOccurrence]
):
    """The model for a full `tryLiquidProbe` command."""

    commandType: TryLiquidProbeCommandType = "tryLiquidProbe"
    params: TryLiquidProbeParams
    result: Optional[TryLiquidProbeResult]

    _ImplementationCls: Type[
        TryLiquidProbeImplementation
    ] = TryLiquidProbeImplementation


class LiquidProbeCreate(BaseCommandCreate[LiquidProbeParams]):
    """The request model for a `liquidProbe` command."""

    commandType: LiquidProbeCommandType = "liquidProbe"
    params: LiquidProbeParams

    _CommandCls: Type[LiquidProbe] = LiquidProbe


class TryLiquidProbeCreate(BaseCommandCreate[TryLiquidProbeParams]):
    """The request model for a `tryLiquidProbe` command."""

    commandType: TryLiquidProbeCommandType = "tryLiquidProbe"
    params: TryLiquidProbeParams

    _CommandCls: Type[TryLiquidProbe] = TryLiquidProbe



"""Temperature module sub-state."""

from dataclasses import dataclass
from typing import NewType, Optional

from opentrons.protocol_engine.types import TemperatureRange
from opentrons.protocol_engine.errors import (
    InvalidTargetTemperatureError,
    NoTargetTemperatureSetError,
)

TemperatureModuleId = NewType("TemperatureModuleId", str)

# TODO (spp, 2022-03-22): Move these values to temperature module definition.
TEMP_MODULE_TEMPERATURE_RANGE = TemperatureRange(min=-9, max=99)


@dataclass(frozen=True)
class TemperatureModuleSubState:
    """Temperature Module specific state.

    Provides calculations and read-only state access
    for an individual loaded Temperaute Module.
    """

    module_id: TemperatureModuleId
    plate_target_temperature: Optional[float]

    @staticmethod
    def validate_target_temperature(celsius: float) -> int:
        """Verify target temperature is within range and of correct type."""
        # Both Gen1 & 2 Temperature modules use 0 decimal precision.
        celsius_int = int(round(celsius, 0))
        if (
            TEMP_MODULE_TEMPERATURE_RANGE.min
            <= celsius_int
            <= TEMP_MODULE_TEMPERATURE_RANGE.max
        ):
            return celsius_int
        else:
            raise InvalidTargetTemperatureError(
                f"Temperature module got an invalid temperature {celsius} °C."
                f" Valid range is {TEMP_MODULE_TEMPERATURE_RANGE}."
            )

    def get_plate_target_temperature(self) -> float:
        """Get the module's target plate temperature."""
        target = self.plate_target_temperature

        if target is None:
            raise NoTargetTemperatureSetError(
                f"Module {self.module_id} does not have a target temperature set."
            )
        return target



"""Command models to deactivate shaker for the Heater-Shaker Module."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler

DeactivateShakerCommandType = Literal["heaterShaker/deactivateShaker"]


class DeactivateShakerParams(BaseModel):
    """Input parameters to deactivate shaker for a Heater-Shaker Module."""

    moduleId: str = Field(..., description="Unique ID of the Heater-Shaker Module.")


class DeactivateShakerResult(BaseModel):
    """Result data from deactivating shaker for a Heater-Shaker."""


class DeactivateShakerImpl(
    AbstractCommandImpl[DeactivateShakerParams, SuccessData[DeactivateShakerResult]]
):
    """Execution implementation of a Heater-Shaker's deactivate shaker command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self, params: DeactivateShakerParams
    ) -> SuccessData[DeactivateShakerResult]:
        """Deactivate shaker for a Heater-Shaker."""
        # Allow propagation of ModuleNotLoadedError and WrongModuleTypeError.
        hs_module_substate = self._state_view.modules.get_heater_shaker_module_substate(
            module_id=params.moduleId
        )

        hs_module_substate.raise_if_labware_latch_not_closed()

        # Allow propagation of ModuleNotAttachedError.
        hs_hardware_module = self._equipment.get_module_hardware_api(
            hs_module_substate.module_id
        )

        if hs_hardware_module is not None:
            await hs_hardware_module.deactivate_shaker()

        return SuccessData(
            public=DeactivateShakerResult(),
        )


class DeactivateShaker(
    BaseCommand[DeactivateShakerParams, DeactivateShakerResult, ErrorOccurrence]
):
    """A command to deactivate shaker for a Heater-Shaker."""

    commandType: DeactivateShakerCommandType = "heaterShaker/deactivateShaker"
    params: DeactivateShakerParams
    result: Optional[DeactivateShakerResult]

    _ImplementationCls: Type[DeactivateShakerImpl] = DeactivateShakerImpl


class DeactivateShakerCreate(BaseCommandCreate[DeactivateShakerParams]):
    """A request to create a Heater-Shaker's deactivate shaker command."""

    commandType: DeactivateShakerCommandType = "heaterShaker/deactivateShaker"
    params: DeactivateShakerParams

    _CommandCls: Type[DeactivateShaker] = DeactivateShaker



"""Calibrate-pipette command for OT3 hardware. request, result, and implementation models."""
from typing import Optional, Type
from typing_extensions import Literal
from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence
from ...types import InstrumentOffsetVector

from opentrons.protocol_engine.resources.ot3_validation import ensure_ot3_hardware


from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.types import OT3Mount
from opentrons.hardware_control import ot3_calibration as calibration

from opentrons.types import MountType

CalibratePipetteCommandType = Literal["calibration/calibratePipette"]


class CalibratePipetteParams(BaseModel):
    """Payload required to calibrate-pipette."""

    mount: MountType = Field(..., description="Instrument mount to calibrate.")


class CalibratePipetteResult(BaseModel):
    """Result data from the execution of a calibrate-pipette command."""

    pipetteOffset: InstrumentOffsetVector = Field(
        ..., description="Offset of calibrated pipette."
    )


class CalibratePipetteImplementation(
    AbstractCommandImpl[CalibratePipetteParams, SuccessData[CalibratePipetteResult]]
):
    """CalibratePipette command implementation."""

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        **kwargs: object,
    ) -> None:
        self._hardware_api = hardware_api

    async def execute(
        self, params: CalibratePipetteParams
    ) -> SuccessData[CalibratePipetteResult]:
        """Execute calibrate-pipette command."""
        # TODO (tz, 20-9-22): Add a better solution to determine if a command can be executed on an OT-3/OT-2
        ot3_api = ensure_ot3_hardware(
            self._hardware_api,
        )
        ot3_mount = OT3Mount.from_mount(params.mount)
        assert (
            ot3_mount is not OT3Mount.GRIPPER
        ), "Expected a Pipette mount but Gripper mount was provided."

        pipette_offset = await calibration.find_pipette_offset(
            hcapi=ot3_api, mount=ot3_mount, slot=5
        )

        await ot3_api.save_instrument_offset(mount=ot3_mount, delta=pipette_offset)

        return SuccessData(
            public=CalibratePipetteResult.construct(
                pipetteOffset=InstrumentOffsetVector.construct(
                    x=pipette_offset.x, y=pipette_offset.y, z=pipette_offset.z
                )
            ),
        )


class CalibratePipette(
    BaseCommand[CalibratePipetteParams, CalibratePipetteResult, ErrorOccurrence]
):
    """Calibrate-pipette command model."""

    commandType: CalibratePipetteCommandType = "calibration/calibratePipette"
    params: CalibratePipetteParams
    result: Optional[CalibratePipetteResult]

    _ImplementationCls: Type[
        CalibratePipetteImplementation
    ] = CalibratePipetteImplementation


class CalibratePipetteCreate(BaseCommandCreate[CalibratePipetteParams]):
    """Create calibrate-pipette command request model."""

    commandType: CalibratePipetteCommandType = "calibration/calibratePipette"
    params: CalibratePipetteParams

    _CommandCls: Type[CalibratePipette] = CalibratePipette



"""Drop tip command request, result, and implementation models."""
from __future__ import annotations

from pydantic import Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from opentrons.protocol_engine.errors.exceptions import TipAttachedError
from opentrons.protocol_engine.resources.model_utils import ModelUtils

from ..state import update_types
from ..types import DropTipWellLocation, DeckPoint
from .pipetting_common import (
    PipetteIdMixin,
    DestinationPositionResult,
    TipPhysicallyAttachedError,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from ..errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from ..state.state import StateView
    from ..execution import MovementHandler, TipHandler


DropTipCommandType = Literal["dropTip"]


class DropTipParams(PipetteIdMixin):
    """Payload required to drop a tip in a specific well."""

    labwareId: str = Field(..., description="Identifier of labware to use.")
    wellName: str = Field(..., description="Name of well to use in labware.")
    wellLocation: DropTipWellLocation = Field(
        default_factory=DropTipWellLocation,
        description="Relative well location at which to drop the tip.",
    )
    homeAfter: Optional[bool] = Field(
        None,
        description=(
            "Whether to home this pipette's plunger after dropping the tip."
            " You should normally leave this unspecified to let the robot choose"
            " a safe default depending on its hardware."
        ),
    )
    alternateDropLocation: Optional[bool] = Field(
        False,
        description=(
            "Whether to alternate location where tip is dropped within the labware."
            " If True, this command will ignore the wellLocation provided and alternate"
            " between dropping tips at two predetermined locations inside the specified"
            " labware well."
            " If False, the tip will be dropped at the top center of the well."
        ),
    )


class DropTipResult(DestinationPositionResult):
    """Result data from the execution of a DropTip command."""

    pass


_ExecuteReturn = (
    SuccessData[DropTipResult] | DefinedErrorData[TipPhysicallyAttachedError]
)


class DropTipImplementation(AbstractCommandImpl[DropTipParams, _ExecuteReturn]):
    """Drop tip command implementation."""

    def __init__(
        self,
        state_view: StateView,
        tip_handler: TipHandler,
        movement: MovementHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._tip_handler = tip_handler
        self._movement_handler = movement
        self._model_utils = model_utils

    async def execute(self, params: DropTipParams) -> _ExecuteReturn:
        """Move to and drop a tip using the requested pipette."""
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName
        home_after = params.homeAfter

        state_update = update_types.StateUpdate()

        if params.alternateDropLocation:
            well_location = self._state_view.geometry.get_next_tip_drop_location(
                labware_id=labware_id,
                well_name=well_name,
                pipette_id=pipette_id,
            )
        else:
            well_location = params.wellLocation

        is_partially_configured = self._state_view.pipettes.get_is_partially_configured(
            pipette_id=pipette_id
        )
        tip_drop_location = self._state_view.geometry.get_checked_tip_drop_location(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_location=well_location,
            partially_configured=is_partially_configured,
        )

        position = await self._movement_handler.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=tip_drop_location,
        )
        deck_point = DeckPoint.construct(x=position.x, y=position.y, z=position.z)
        state_update.set_pipette_location(
            pipette_id=pipette_id,
            new_labware_id=labware_id,
            new_well_name=well_name,
            new_deck_point=deck_point,
        )

        try:
            await self._tip_handler.drop_tip(
                pipette_id=pipette_id, home_after=home_after
            )
        except TipAttachedError as exception:
            error = TipPhysicallyAttachedError(
                id=self._model_utils.generate_id(),
                createdAt=self._model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=self._model_utils.generate_id(),
                        createdAt=self._model_utils.get_timestamp(),
                        error=exception,
                    )
                ],
                errorInfo={"retryLocation": position},
            )
            state_update_if_false_positive = update_types.StateUpdate()
            state_update_if_false_positive.update_pipette_tip_state(
                pipette_id=params.pipetteId, tip_geometry=None
            )
            return DefinedErrorData(
                public=error,
                state_update=state_update,
                state_update_if_false_positive=state_update_if_false_positive,
            )
        else:
            state_update.update_pipette_tip_state(
                pipette_id=params.pipetteId, tip_geometry=None
            )
            return SuccessData(
                public=DropTipResult(position=deck_point),
                state_update=state_update,
            )


class DropTip(BaseCommand[DropTipParams, DropTipResult, TipPhysicallyAttachedError]):
    """Drop tip command model."""

    commandType: DropTipCommandType = "dropTip"
    params: DropTipParams
    result: Optional[DropTipResult]

    _ImplementationCls: Type[DropTipImplementation] = DropTipImplementation


class DropTipCreate(BaseCommandCreate[DropTipParams]):
    """Drop tip command creation request model."""

    commandType: DropTipCommandType = "dropTip"
    params: DropTipParams

    _CommandCls: Type[DropTip] = DropTip



from . import (
    models,
    gripper_offset as gripper_offset,
)
from .pipette_offset import (
    save_pipette_calibration,
    clear_pipette_offset_calibrations,
    get_pipette_offset,
    delete_pipette_offset_file,
)
from .module_offset import (
    save_module_calibration,
    clear_module_offset_calibrations,
    get_module_offset,
    delete_module_offset_file,
    load_all_module_offsets,
)

__all__ = [
    "models",
    "gripper_offset",
    "save_pipette_calibration",
    "clear_pipette_offset_calibrations",
    "get_pipette_offset",
    "delete_pipette_offset_file",
    "save_module_calibration",
    "clear_module_offset_calibrations",
    "get_module_offset",
    "delete_module_offset_file",
    "load_all_module_offsets",
]



""" opentrons.protocol_api.module_geometry: classes and functions for modules
as deck objects

This module provides things like :py:class:`ModuleGeometry` and
:py:func:`load_module` to create and manipulate module objects as geometric
objects on the deck (as opposed to calling commands on them, which is handled
by :py:mod:`.module_contexts`)
"""
from __future__ import annotations

import logging
from enum import Enum
from typing import TYPE_CHECKING, Optional

import numpy as np
from numpy.typing import NDArray

from opentrons_shared_data import module
from opentrons_shared_data.module.types import ModuleDefinitionV3
from opentrons_shared_data.module import OLD_TC_GEN2_LABWARE_OFFSET

from opentrons.types import Location, Point, LocationLabware
from opentrons.motion_planning.adjacent_slots_getters import (
    get_north_south_slots,
    get_east_west_slots,
    get_adjacent_slots,
)

from opentrons.drivers.types import ThermocyclerLidStatus
from opentrons.hardware_control.modules.types import (
    ModuleModel,
    ModuleType,
    module_model_from_string,
    ThermocyclerModuleModel,
)

if TYPE_CHECKING:
    from opentrons.protocol_api.labware import Labware


_log = logging.getLogger(__name__)


class ThermocyclerConfiguration(str, Enum):
    FULL = "full"
    SEMI = "semi"


class NoSuchModuleError(ValueError):
    """An error raised if a requested model does not match a known module."""


class PipetteMovementRestrictedByHeaterShakerError(Exception):
    """Error raised when trying to move to labware restricted by Heater-Shaker."""


# TODO (spp, 2022-05-09): add tests
class ModuleGeometry:
    """
    This class represents an active peripheral, such as an Opentrons Magnetic
    Module, Temperature Module or Thermocycler Module. It defines the physical
    geometry of the device (primarily the offset that modifies the position of
    the labware mounted on top of it).
    """

    @property
    def separate_calibration(self) -> bool:
        _log.warning(
            "ModuleGeometry.separate_calibrations is a deprecated internal property."
            " It has no longer has meaning, but will always return `True`"
        )
        return True

    def __init__(
        self,
        display_name: str,
        model: ModuleModel,
        module_type: ModuleType,
        offset: Point,
        overall_height: float,
        height_over_labware: float,
        parent: Location,
    ) -> None:
        """
        Create a Module for tracking the position of a module.

        Note that modules do not currently have a concept of calibration apart
        from calibration of labware on top of the module. The practical result
        of this is that if the module parent :py:class:`.Location` is
        incorrect, then a correct calibration of one labware on the deck would
        be incorrect on the module, and vice-versa. Currently, the way around
        this would be to correct the :py:class:`.Location` so that the
        calibrated labware is targeted accurately in both positions.

        :param display_name: A human-readable display name of only the module
                             (for instance, "Thermocycler Module" - not
                             including parents or location)
        :param model: The model of module this represents
        :param offset: The offset from the slot origin at which labware loaded
                       on this module should be placed
        :param overall_height: The height of the module without labware
        :param height_over_labware: The height of this module over the top of
                                    the labware
        :param parent: A location representing location of the front left of
                       the outside of the module (usually the front-left corner
                       of a slot on the deck).
        :type parent: :py:class:`.Location`
        """
        self._parent = parent
        self._module_type = module_type

        self._display_name = display_name
        self._model = model
        self._offset = offset
        self._height = overall_height + self._parent.point.z
        self._over_labware = height_over_labware
        self._labware: Optional[Labware] = None
        self._location = Location(point=self._offset + self._parent.point, labware=self)

    def add_labware(self, labware: Labware) -> Labware:
        if self._labware is not None:
            raise RuntimeError(f"{self._labware} is already on this module")
        self._labware = labware
        return self._labware

    def reset_labware(self) -> None:
        self._labware = None

    @property
    def display_name(self) -> str:
        return self._display_name

    @property
    def model(self) -> ModuleModel:
        return self._model

    @property
    def load_name(self) -> str:
        # Mypy (at the time of writing, v0.910) incorrectly types self.model.value as
        # Any. It seems to have trouble with it being a union of enums.
        assert isinstance(self.model.value, str)
        return self.model.value

    @property
    def module_type(self) -> ModuleType:
        """The Moduletype"""
        return self._module_type

    @property
    def parent(self) -> LocationLabware:
        return self._parent.labware.object

    @property
    def labware(self) -> Optional[Labware]:
        return self._labware

    @property
    def location(self) -> Location:
        """
        :return: a :py:class:`.Location` representing the top of the module
        """
        return self._location

    @property
    def labware_offset(self) -> Point:
        """
        :return: a :py:class:`.Point` representing the transformation
        between the critical point of the module and the critical
        point of its contained labware
        """
        return self._offset

    @property
    def highest_z(self) -> float:
        if self.labware:
            return self.labware.highest_z + self._over_labware
        else:
            return self._height

    def __repr__(self) -> str:
        location = f" on {self.parent}" if isinstance(self.parent, str) else ""
        return f"{self._display_name}{location}"


class ThermocyclerGeometry(ModuleGeometry):
    def __init__(
        self,
        display_name: str,
        model: ModuleModel,
        module_type: ModuleType,
        offset: Point,
        overall_height: float,
        height_over_labware: float,
        lid_height: float,
        parent: Location,
        configuration: ThermocyclerConfiguration = ThermocyclerConfiguration.FULL,
    ) -> None:
        """
        Create a Module for tracking the position of a module.

        Note that modules do not currently have a concept of calibration apart
        from calibration of labware on top of the module. The practical result
        of this is that if the module parent :py:class:`.Location` is
        incorrect, then a correct calibration of one labware on the deck would
        be incorrect on the module, and vice-versa. Currently, the way around
        this would be to correct the :py:class:`.Location` so that the
        calibrated labware is targeted accurately in both positions.

        :param display_name: A human-readable display name of only the module
                             (for instance, "Thermocycler Module" - not
                             including parents or location)
        :param model: The model of module this represents
        :param offset: The offset from the slot origin at which labware loaded
                       on this module should be placed
        :param overall_height: The height of the module without labware
        :param height_over_labware: The height of this module over the top of
                                    the labware
        :param parent: A location representing location of the front left of
                       the outside of the module (usually the front-left corner
                       of a slot on the deck).
        :type parent: :py:class:`.Location`
        :param configuration: Used to specify the slot configuration of
                              the Thermocycler. It should by of type
                              :py:class:`.ThermocyclerConfiguration` and can
                              either be FULL or SEMI.
        """
        super().__init__(
            display_name,
            model,
            module_type,
            offset,
            overall_height,
            height_over_labware,
            parent,
        )
        self._lid_height = lid_height
        self._lid_status = ThermocyclerLidStatus.OPEN  # Needs to reflect true status
        self._configuration = configuration
        if self.is_semi_configuration:
            self._offset = self._offset + Point(-23.28, 0, 0)

    @property
    def highest_z(self) -> float:
        # TODO: BC 2019-08-27 this highest_z value represents the distance
        # from the top of the open TC chassis to the base. Once we have a
        # more robust collision detection system in place, the collision
        # model for the TC should change based on it's lid_status
        # (open or closed). A prerequisite for that check will be
        # path-specific highest z calculations, as opposed to the current
        # global check on instrument.move_to. For example: a move from slot 1
        # to slot 3 should only check the highest z of all deck items between
        # the source and destination in the x,y plane.
        return super().highest_z

    @property
    def lid_status(self) -> ThermocyclerLidStatus:
        return self._lid_status

    @lid_status.setter
    def lid_status(self, status: ThermocyclerLidStatus) -> None:
        self._lid_status = status

    @property
    def is_semi_configuration(self) -> bool:
        return bool(self._configuration == ThermocyclerConfiguration.SEMI)

    # TODO(mc, 2022-11-16): this method causes bugs and should not be used;
    # Thermocycler `configuration="semi"` does not work properly and should be removed
    # https://opentrons.atlassian.net/browse/RSS-106
    def labware_accessor(self, labware: Labware) -> Labware:
        from opentrons.protocol_api.labware import Labware
        from opentrons.protocol_api.core.legacy.legacy_labware_core import (
            LegacyLabwareCore,
        )

        # Block first three columns from being accessed
        definition = labware._core.get_definition()
        definition["ordering"] = definition["ordering"][2::]
        return Labware(
            core=LegacyLabwareCore(definition, super().location),
            api_version=labware.api_version,
            protocol_core=None,  # type: ignore[arg-type]
            core_map=None,  # type: ignore[arg-type]
        )

    def add_labware(self, labware: Labware) -> Labware:
        if self._labware is not None:
            raise RuntimeError(f"{self._labware} is already on this module")
        if self.lid_status == "closed":
            raise RuntimeError("Cannot place labware in closed module")
        if self.is_semi_configuration:
            self._labware = self.labware_accessor(labware)
        else:
            self._labware = labware
        return self._labware

    def flag_unsafe_move(
        self,
        to_loc: Location,
        from_loc: Location,
        lid_position: Optional[str],
    ) -> None:
        to_lw, to_well = to_loc.labware.get_parent_labware_and_well()
        from_lw, from_well = from_loc.labware.get_parent_labware_and_well()

        if (
            self.labware is not None
            and (self.labware == to_lw or self.labware == from_lw)
            and lid_position != "open"
        ):
            raise RuntimeError(
                "Cannot move to labware loaded in Thermocycler"
                " when lid is not fully open."
            )


class HeaterShakerGeometry(ModuleGeometry):
    """Class holding the state of a Heater-Shaker's physical geometry."""

    def flag_unsafe_move(
        self,
        to_slot: int,
        is_tiprack: bool,
        is_using_multichannel: bool,
        is_labware_latch_closed: bool,
        is_plate_shaking: bool,
    ) -> None:
        """Raise error if unsafe to move pipette due to Heater-Shaker placement."""
        assert isinstance(self.parent, str), "Could not determine module slot location"

        heater_shaker_slot = int(self.parent)
        dest_east_west = to_slot in get_east_west_slots(heater_shaker_slot)
        dest_north_south = to_slot in get_north_south_slots(heater_shaker_slot)
        dest_heater_shaker = to_slot == heater_shaker_slot

        # If Heater-Shaker is running, can't move to or around it
        if (
            any([dest_east_west, dest_north_south, dest_heater_shaker])
            and is_plate_shaking
        ):
            raise PipetteMovementRestrictedByHeaterShakerError(
                "Cannot move pipette to Heater-Shaker or adjacent slot while module is shaking"
            )

        # If Heater-Shaker's latch is open, can't move to it or east and west of it
        elif (dest_east_west or dest_heater_shaker) and not is_labware_latch_closed:
            raise PipetteMovementRestrictedByHeaterShakerError(
                "Cannot move pipette to Heater-Shaker or adjacent slot to the left or right while labware latch is open"
            )

        elif is_using_multichannel:
            # Can't go to east/west slot under any circumstances if pipette is multi-channel
            if dest_east_west:
                raise PipetteMovementRestrictedByHeaterShakerError(
                    "Cannot move 8-Channel pipette to slot adjacent to the left or right of Heater-Shaker"
                )
            # Can only go north/south if the labware is a tip rack
            elif dest_north_south and not is_tiprack:
                raise PipetteMovementRestrictedByHeaterShakerError(
                    "Cannot move 8-Channel pipette to non-tip-rack labware directly in front of or behind a Heater-Shaker"
                )

    def is_pipette_blocking_shake_movement(
        self, pipette_location: Optional[Location]
    ) -> bool:
        """Check whether pipette is parked adjacent to Heater-Shaker.

        Returns True if pipette's last known location was in a slot adjacent to or
        on the Heater-Shaker. Also returns True if last location is not known or is
        not associated with a slot.
        """
        if pipette_location is None:
            # If we don't know the pipette's latest location then let's be extra
            # cautious and call it blocking
            return True

        pipette_location_slot = pipette_location.labware.first_parent()
        if pipette_location_slot is None:
            # If a location is not associated w/ a slot (e.g., if it has labware=None)
            # then we don't know if it's close to the h/s, so, we will be cautious
            # and call it blocking
            return True

        heater_shaker_slot = self.parent

        assert isinstance(
            heater_shaker_slot, str
        ), "Could not determine module slot location"

        return heater_shaker_slot == pipette_location_slot or int(
            pipette_location_slot
        ) in get_adjacent_slots(int(heater_shaker_slot))

    def is_pipette_blocking_latch_movement(
        self, pipette_location: Optional[Location]
    ) -> bool:
        """Check whether pipette is parked left or right of Heater-Shaker.

        Returns True is pipette's last known location was on the Heater-Shaker
        or in a slot adjacent to its left or right. Also returns True if last
        location is not known or is not associated with a slot.
        """
        if pipette_location is None:
            # If we don't know the pipette's latest location then let's be extra
            # cautious and call it blocking
            return True

        pipette_location_slot = pipette_location.labware.first_parent()
        if pipette_location_slot is None:
            # If a location is not associated w/ a slot (e.g., if it has labware=None)
            # then we don't know if it's close to the h/s, so, we will be cautious
            # and call it blocking
            return True

        heater_shaker_slot = self.parent

        assert isinstance(
            heater_shaker_slot, str
        ), "Could not determine module slot location"

        return heater_shaker_slot == pipette_location_slot or int(
            pipette_location_slot
        ) in get_east_west_slots(int(heater_shaker_slot))


def create_geometry(
    definition: ModuleDefinitionV3,
    parent: Location,
    configuration: Optional[str],
) -> ModuleGeometry:
    """Create a module geometry object from the module's definition and location.

    The definition should be schema checked before being passed to this
    function; all definitions passed here are assumed to be valid.
    """
    module_type = ModuleType(definition["moduleType"])
    module_model = module_model_from_string(definition["model"])
    overall_height = definition["dimensions"]["bareOverallHeight"]
    height_over_labware = definition["dimensions"]["overLabwareHeight"]
    display_name = definition["displayName"]

    if module_model == ThermocyclerModuleModel.THERMOCYCLER_V2:
        pre_transform: NDArray[np.double] = np.array(
            (
                OLD_TC_GEN2_LABWARE_OFFSET["x"],
                OLD_TC_GEN2_LABWARE_OFFSET["y"],
                OLD_TC_GEN2_LABWARE_OFFSET["z"],
                1,
            )
        )
    else:
        pre_transform = np.array(
            (
                definition["labwareOffset"]["x"],
                definition["labwareOffset"]["y"],
                definition["labwareOffset"]["z"],
                1,
            )
        )
    if not parent.labware.is_slot:
        par = ""
        _log.warning(
            f"module location parent labware was {parent.labware} which is"
            "not a slot name; slot transforms will not be loaded"
        )
    else:
        par = str(parent.labware.object)

    # This ModuleGeometry class is only used in Python Protocol API versions <=2.13,
    # which only support the OT-2, so we can ignore OT-3s and hard-code "ot2_standard."
    # We also assume that "ot2_short_trash" has the same transforms as "ot2_standard."
    xforms_ser = (
        definition["slotTransforms"]
        .get("ot2_standard", {})
        .get(
            par,
            {"labwareOffset": [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]},
        )
    )
    xform_ser = xforms_ser["labwareOffset"]

    # apply the slot transform if any
    xform: NDArray[np.double] = np.array(xform_ser)
    xformed = np.dot(xform, pre_transform)
    labware_offset = Point(xformed[0], xformed[1], xformed[2])

    if module_type == ModuleType.MAGNETIC or module_type == ModuleType.TEMPERATURE:
        return ModuleGeometry(
            parent=parent,
            offset=labware_offset,
            overall_height=overall_height,
            height_over_labware=height_over_labware,
            model=module_model,
            module_type=module_type,
            display_name=display_name,
        )
    elif module_type == ModuleType.THERMOCYCLER:
        return ThermocyclerGeometry(
            parent=parent,
            offset=labware_offset,
            overall_height=overall_height,
            height_over_labware=height_over_labware,
            model=module_model,
            module_type=module_type,
            display_name=display_name,
            lid_height=definition["dimensions"]["lidHeight"],
            configuration=(
                ThermocyclerConfiguration(configuration)
                if configuration is not None
                else ThermocyclerConfiguration.FULL
            ),
        )
    elif module_type == ModuleType.HEATER_SHAKER:
        return HeaterShakerGeometry(
            parent=parent,
            offset=labware_offset,
            overall_height=overall_height,
            height_over_labware=height_over_labware,
            model=module_model,
            module_type=module_type,
            display_name=display_name,
        )
    else:
        raise AssertionError(f"Module type {module_type} is invalid")


def load_definition(model: str) -> ModuleDefinitionV3:
    """Load the geometry definition of a given module by model."""
    try:
        return module.load_definition("3", model)
    except module.ModuleNotFoundError:
        raise NoSuchModuleError(f'Could not find a module with model "{model}"')


def models_compatible(
    requested_model: ModuleModel, candidate_definition: ModuleDefinitionV3
) -> bool:
    """Check if a requested model is compatible with a given definition."""
    return (
        requested_model.value == candidate_definition["model"]
        or requested_model.value in candidate_definition["compatibleWith"]
    )



"""Command models to close a Thermocycler's lid."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence
from ...state import update_types
from opentrons.protocol_engine.types import MotorAxis

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler, MovementHandler


CloseLidCommandType = Literal["thermocycler/closeLid"]


class CloseLidParams(BaseModel):
    """Input parameters to close a Thermocycler's lid."""

    moduleId: str = Field(..., description="Unique ID of the Thermocycler.")


class CloseLidResult(BaseModel):
    """Result data from closing a Thermocycler's lid."""


class CloseLidImpl(AbstractCommandImpl[CloseLidParams, SuccessData[CloseLidResult]]):
    """Execution implementation of a Thermocycler's close lid command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        movement: MovementHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._movement = movement

    async def execute(self, params: CloseLidParams) -> SuccessData[CloseLidResult]:
        """Close a Thermocycler's lid."""
        state_update = update_types.StateUpdate()

        thermocycler_state = self._state_view.modules.get_thermocycler_module_substate(
            params.moduleId
        )
        thermocycler_hardware = self._equipment.get_module_hardware_api(
            thermocycler_state.module_id
        )

        # move the pipettes and gantry over the trash
        # do not home plunger axes because pipettes may be holding liquid
        axes_to_home = [
            MotorAxis.X,
            MotorAxis.Y,
        ] + self._state_view.motion.get_robot_mount_axes()
        await self._movement.home(axes=axes_to_home)
        state_update.clear_all_pipette_locations()

        if thermocycler_hardware is not None:
            await thermocycler_hardware.close()

        return SuccessData(public=CloseLidResult(), state_update=state_update)


class CloseLid(BaseCommand[CloseLidParams, CloseLidResult, ErrorOccurrence]):
    """A command to close a Thermocycler's lid."""

    commandType: CloseLidCommandType = "thermocycler/closeLid"
    params: CloseLidParams
    result: Optional[CloseLidResult]

    _ImplementationCls: Type[CloseLidImpl] = CloseLidImpl


class CloseLidCreate(BaseCommandCreate[CloseLidParams]):
    """A request to close a Thermocycler's lid."""

    commandType: CloseLidCommandType = "thermocycler/closeLid"
    params: CloseLidParams

    _CommandCls: Type[CloseLid] = CloseLid
